from typing import Union, Optional, Dict, List
import uuid

import numpy as np

root_datatypes = ["bool", "Bool_t", "Byte_t", "char", "char*", "Char_t", 
                  "double", "Double32_t", "Double_t", "float",
                  "Float16_t", "Float_t", "int", "Int_t", 
                  "long", "long long", "Long_t", "Long64_t",
                  "short", "Short_t", "Size_t", "UChar_t",
                  "UInt_t", "ULong64_t", "ULong_t",
                  "unsigned", "unsigned char", "unsigned int",
                  "unsigned long", "unsigned long long",
                  "unsigned short", "UShort_t"]

uproot_datatypes = ["double", "float", "int", "int64_t", "char*", "int32_t", "uint64_t"]

def get_default_library(custom_columns:bool=False):
    if custom_columns:
        return "root"
    try:
        import uproot
        has_uproot = True
        from packaging import version
        uproot_version = uproot.__version__
        if version.parse(uproot_version) < version.parse("4.2.0"):
            print("WARNING: uproot version too old (<4.2.0), will switch to using ROOT instead")
            has_uproot = False
    except ImportError:
        has_uproot = False
    if has_uproot:
        return "uproot"
    return "root"

def downcast_dataframe(df):
    import pandas as pd
    fcols = df.select_dtypes('float').columns
    icols = df.select_dtypes('integer').columns

    df[fcols] = df[fcols].apply(pd.to_numeric, downcast='float')
    df[icols] = df[icols].apply(pd.to_numeric, downcast='integer')

def array2root(array_data:Dict[str, np.ndarray], fname:str, treename:str,
               library:str="auto", multithread:bool=True):
    if library.lower() == "auto":
        library = get_default_library()
    if library == "root":
        from quickstats.interface.root.helper import RMultithreadEnv
        from quickstats.interface.cppyy.vectorize import np_type_str_maps
        with RMultithreadEnv(multithread):
            columns = list(array_data.keys())
            snapshot_templates = []
            for column in columns:
                template_type = np_type_str_maps.get(array_data[column].dtype, None)
                if template_type is None:
                    raise ValueError(f"unsupported array type \"{array_data[column].dtype}\""
                                     f" from the column \"{column}\"")
                if template_type == "bool":
                    template_type = "int"
                    array_data[column] = array_data[column].astype("int32")
                snapshot_templates.append(template_type)
            snapshot_templates = tuple(snapshot_templates)
            import ROOT
            df = ROOT.RDF.MakeNumpyDataFrame(array_data)
            df.Snapshot.__getitem__(snapshot_templates)(treename, fname, columns)
    elif library == "uproot":
        import uproot
        from packaging import version
        uproot_version = uproot.__version__
        if version.parse(uproot_version) < version.parse("4.2.0"):
            raise RuntimeError("uproot version too old (requires 4.2.0+)")
        file = uproot.recreate(fname)
        file[treename] = array_data
        file.close()
    else:
        raise RuntimeError(f'unknown library "{library}" for root data conversion')            
        
numpy2root = array2root

def dataframe2numpy(df:"pandas.DataFrame", columns:Optional[List[str]]=None):
    if columns is not None:
        arrays = dict(zip(columns, df[columns].to_numpy().T))
    else:
        arrays = dict(zip(df.columns.values, df.to_numpy().T))
    for column in arrays:
        arrays[column] = arrays[column].astype(df[column].dtype)
    return arrays 

def numpy2dataframe(array_data:Dict[str, np.ndarray]):
    array_shallow_copy = {**array_data}
    for key, array in array_data.items():
        if (array.ndim > 1) and (array.dtype != object):
            array_shallow_copy[key] = list(array)
    import pandas as pd
    df = pd.DataFrame(array_shallow_copy)
    return df

array2dataframe = numpy2dataframe

def dataframe2root(df:"pandas.DataFrame", fname:str, treename:str,
                   columns:Optional[List[str]]=None,
                   library:str="auto", multithread:bool=True):
    array_data = dataframe2numpy(df, columns)
    array2root(array_data, fname, treename, library=library,
               multithread=multithread)
    
def uproot_get_standard_columns(uproot_tree):
    typenames = uproot_tree.typenames()
    columns = list(typenames.keys())
    column_types = list(typenames.values())
    return np.array(columns)[np.where(np.isin(column_types, uproot_datatypes))]

def rdf2numpy(rdf, columns:Union[Dict[str, str], List[str]]=None,
              cut:Optional[str]=None, convert_vectors:bool=True,
              remove_non_standard_types:bool=True):
    if cut is not None:
        rdf = rdf.Filter(cut)     
    rename_columns = {}
    if columns is None:
        save_columns = [str(name) for name in rdf.GetColumnNames()]
    elif isinstance(columns, dict):
        save_columns = []
        for column_name, definition in columns.items():
            if column_name == definition:
                save_columns.append(column_name)
                continue
            if rdf.HasColumn(column_name):
                new_column_name = f"var_{uuid.uuid4().hex}"
                rename_columns[new_column_name] = column_name
                column_name = new_column_name
            rdf = rdf.Define(column_name, definition)
            save_columns.append(column_name)
    else:
        save_columns = list(columns)
    if remove_non_standard_types:
        column_types = np.array([rdf.GetColumnType(column_name) for column_name in save_columns])
        save_columns = list(np.array(save_columns)[np.where(np.isin(column_types, root_datatypes))])
        rename_columns = {k:v for k,v in rename_columns.items() if k in save_columns}
    vector_columns = []
    if convert_vectors:
        vector_columns_tmp = []
        for column_name in save_columns:
            column_type = rdf.GetColumnType(column_name)
            if column_type.count("ROOT::VecOps::RVec") == 1:
                vector_columns_tmp.append(column_name)
        if len(vector_columns_tmp) > 0:
            import quickstats
            quickstats.load_processor_methods()
            for column_name in vector_columns_tmp:
                new_column_name = f"var_{uuid.uuid4().hex}"
                if column_name in save_columns:
                    save_columns = [new_column_name if name == column_name else name for name in save_columns]
                # mapped twice
                if column_name in rename_columns:
                    rename_columns[new_column_name] = rename_columns.pop(column_name)
                else:
                    rename_columns[new_column_name] = column_name
                rdf = rdf.Define(new_column_name, f"RVec2Vec({column_name})")
                vector_columns.append(new_column_name)

    available_columns = [str(name) for name in rdf.GetColumnNames()]
    missing_columns = np.setdiff1d(save_columns, available_columns)
    if len(missing_columns) > 0:
        raise RuntimeError(f'missing column(s): {", ".join(missing_columns)}')
    result = rdf.AsNumpy(save_columns)
    for vector_column in vector_columns:
        # not the most efficient way, but easiest
        numpy_array = np.array([np.array(v.data()) for v in result[vector_column]], dtype=object)
        # in case it't array of regular size
        if len(numpy_array) and (numpy_array[0].dtype == object):
            result[vector_column] = np.array([np.array(v.data()) for v in result[vector_column]])
        else:
            result[vector_column] = numpy_array
    for old_column, new_column in rename_columns.items():
        result[new_column] = result.pop(old_column)
    # reorder the columns to match the order given by the user
    if columns is not None:
        result = {column: result[column] for column in columns if column in result}
    return result
    
def root2numpy(filename:Union[str, List[str]], treename:str,
               columns:Union[Dict[str, str], List[str]]=None,
               cut:Optional[str]=None, convert_vectors:bool=True,
               remove_non_standard_types:bool=True,
               library:str="auto"):
    if library.lower() == "auto":
        library = get_default_library(custom_columns=isinstance(columns,dict))
    if library.lower() == "root":
        import ROOT
        rdf = ROOT.RDataFrame(treename, filename)
        return rdf2numpy(rdf, columns=columns, cut=cut,
                         convert_vectors=convert_vectors,
                         remove_non_standard_types=remove_non_standard_types)
    elif library.lower() == "uproot":
        if isinstance(columns, dict):
            raise RuntimeError('defining new columns are not supported when using "uproot" as the library')
        import uproot
        if isinstance(filename, str):
            f = uproot.open(filename)
            t = f[treename]
            if remove_non_standard_types:
                standard_columns = uproot_get_standard_columns(t)
                if columns is None:
                    columns = standard_columns
                else:
                    columns = [column for column in columns if column in standard_columns]
            return f[treename].arrays(columns, library="numpy", cut=cut)
        else:
            # iterate over multiple files
            files = {f:treename for f in filename}
            if remove_non_standard_types:
                filter_typename = list(uproot_datatypes)
            else:
                filter_typename = None
            result = {}
            for batch in uproot.iterate(files, expressions=columns,
                                        filter_typename=filter_typename,
                                        cut=cut, library="numpy"):
                for column in batch:
                    if column not in result:
                        result[column] = batch[column]
                    else:
                        result[column] = np.concatenate([result[column], batch[column]])
            return result
    else:
        raise RuntimeError(f'unknown library "{library}" for root data conversion')

root2array = root2numpy
        
def root2dataframe(filename:Union[str, List[str]], treename:str,
                   columns:Union[Dict[str, str], List[str]]=None,
                   cut:Optional[str]=None,
                   remove_non_standard_types:bool=True,
                   downcast:bool=True,
                   library:str="auto"):
    if library.lower() == "auto":
        library = get_default_library(custom_columns=isinstance(columns,dict))
    if library.lower() == "root":
        numpy_data = root2numpy(filename, treename, columns=columns, cut=cut,
                                convert_vectors=True,
                                remove_non_standard_types=remove_non_standard_types,
                                library=library)
        result = numpy2dataframe(numpy_data)
    elif library.lower() == "uproot":
        if isinstance(columns, dict):
            raise RuntimeError('defining new columns are not supported when using "uproot" as the library')
        import uproot
        if isinstance(filename, str):
            f = uproot.open(filename)
            t = f[treename]
            if remove_non_standard_types:
                standard_columns = uproot_get_standard_columns(t)
                if columns is None:
                    columns = standard_columns
                else:
                    columns = [column for column in columns if column in standard_columns]
            result = f[treename].arrays(columns, library="pandas")
        else:
            import pandas as pd
            # iterate over multiple files
            files = {f:treename for f in filename}
            if remove_non_standard_types:
                filter_typename = list(uproot_datatypes)
            else:
                filter_typename = None
            result = None
            for batch in uproot.iterate(files, expressions=columns,
                                        filter_typename=filter_typename,
                                        cut=cut, library="pandas"):
                if result is None:
                    result = batch
                else:
                    result = pd.concat([result, batch])
    if downcast:
        downcast_dataframe(result)
    return result

def root2rdataset(filename:Union[str, List[str], "quickstats.PathManager"], treename:str,
                  observable:Union[str, dict, "ROOT.RooRealVar",
                                   "quickstats.interface.root.RooRealVar"],
                  weight_name:Optional[str]=None,
                  dataset_name:str="obsData"):
    from quickstats.components.modelling import TreeDataSource
    source = TreeDataSource(treename, filename, observable, weight_name)
    dataset = source.construct_dataset(dataset_name)
    return dataset

def rdataset2numpy(dataset:"ROOT.RooDataSet"):
    from quickstats.interface.root import RooDataSet
    return RooDataSet.to_numpy(dataset)

def rdataset2dataframe(dataset:"ROOT.RooDataSet"):
    from quickstats.interface.root import RooDataSet
    return RooDataSet.to_pandas(dataset)

def rdataset2hist(dataset:"ROOT.RooDataSet"):
    pass

def root2hist(filename:Union[str, List[str]], treename:str,
              column:str, bins:int=10, range:List[float]=None, weight_column:Optional[str]=None):
    pass


from typing import Optional, Union, List, Dict

import ROOT

from quickstats.utils.common_utils import remove_list_duplicates

from .settings import SELF_SYST_DOMAIN_KEYWORD, PDF_PREFIX

class Sample:
    
    # deprecated name
    @property
    def syst_groups(self):
        return self.syst_domains
    
    @syst_groups.setter
    def syst_groups(self, value):
        self.syst_domains = value
    
    def __init__(self):
        self.process = ""
        self.input_file = ""
        self.norm_factors = []
        self.shape_factors = []
        self.syst_domains = []
        self.expected = ROOT.RooArgSet()
        self.model_name = ""
        self.norm_name = ""
        self.share_pdf_group = ""
        
    def __eq__(self, other:Union[str, "Sample"]):
        if isinstance(other, str):
            return self.process == other
        elif isinstance(other, type(self)):
            return self.process == other.process
        else:
            raise TypeError(f"cannot compare object of type {type(other)} with "
                            "object of type Sample")
        
    def is_equal(self, other:Union[str, "Sample"]):
        return self.__eq__(other)
            
    def get_tag_name(self) -> str:
        is_shared_pdf = self.share_pdf_group != ""
        if is_shared_pdf:
            tag_name = self.share_pdf_group
        else:
            tag_name = self.process
        return tag_name
    
    def update_model_name(self):
        self.model_name = f"{PDF_PREFIX}{self.get_tag_name()}"
        
    def set_syst_domains(self, syst_domains:List[str]):
        # if the domain contains ":self:", then do not take any systematics from outside
        if SELF_SYST_DOMAIN_KEYWORD in syst_domains:
            syst_domains = []
        # all the systematics under the process should be included in any case  
        if self.process not in syst_domains:
            syst_domains.append(self.process)
        self.syst_domains = remove_list_duplicates(syst_domains)


from typing import Optional
import re

from .rooproc_rdf_action import RooProcRDFAction

class RooProcFilter(RooProcRDFAction):
    
    def __init__(self, expression:str, name:Optional[str]=None):
        super().__init__(expression=expression,
                         name=name)
        
    @classmethod
    def parse(cls, main_text:str, block_text:Optional[str]=None):
        name_literals = re.findall(r"@{([^{}]+)}", main_text)
        if len(name_literals) == 0:
            name = None
            expression = main_text.strip()
        elif len(name_literals) == 1:
            name = name_literals[0]
            expression = main_text.replace("@{" + name + "}", "").strip()
        else:
            raise RuntimeError(f"multiple filter names detected in the expression `{main_text}`")
        return cls(name=name, expression=expression)
        
    def _execute(self, rdf:"ROOT.RDataFrame", **params):
        expression = params['expression']
        name = params.get("name", None)
        if name is not None:
            rdf_next = rdf.Filter(expression, name)
        else:
            rdf_next = rdf.Filter(expression)
        return rdf_next


from typing import Optional
import re

from .rooproc_rdf_action import RooProcRDFAction

class RooProcAlias(RooProcRDFAction):
    
    def __init__(self, alias:str, column_name:str):
        super().__init__(alias=alias, column_name=column_name)
        
    @classmethod
    def parse(cls, main_text:str, block_text:Optional[str]=None):
        result = re.search(r"^\s*(\w+)\s*=\s*([\w\.]+)\s*$", main_text)
        if not result:
            if re.search(r"^\s*(\w+)\s*=\s*([\w\.]+)", main_text):
                raise RuntimeError(f"can not alias an expression, please use DEFINE instead")
            raise RuntimeError(f"invalid expression {main_text}")
        alias = result.group(1)
        column_name = result.group(2)
        return cls(alias=alias, column_name=column_name)        
        
    def _execute(self, rdf:"ROOT.RDataFrame", **params):
        alias = params['alias']
        column_name = params['column_name']
        rdf_next = rdf.Alias(alias, column_name)
        return rdf_next


from typing import Optional, Union, List, Dict
import os
import re
import time
import difflib
from functools import partial

import numpy as np

import quickstats
from quickstats import semistaticmethod, AbstractObject, GeneralEnum, PathManager
from quickstats.utils.io import Verbosity
from quickstats.utils.xml_tools import TXMLTree
from quickstats.maths.numerics import to_bool
from quickstats.utils.root_utils import load_macro, get_macro_TClass
from quickstats.utils.common_utils import set_unlimited_stacksize

import ROOT

class WSObjectType(GeneralEnum):
    DEFINED    = 0
    FUNCTION   = 1
    VARIABLE   = 2
    CONSTRAINT = 3

class XMLAttribType(GeneralEnum):
    STR        = 0
    BOOL       = 1
    INT        = 2
    FLOAT      = 3
    STR_LIST   = 4
    INT_LIST   = 5
    FLOAT_LIST = 6

class BracketType(GeneralEnum):
    ROUND  = (1, r"(.+)\((.+)\)")
    SQUARE = (2, r"(.+)\[(.+)\]")
    CURLY  = (3, r"(.+){(.+)}")
    ANGLE  = (4, r"(.+)<(.+)>")
    
    def __new__(cls, value:int, regex:str):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.regex = re.compile(regex)
        return obj
    
ConversionOperators = {
    XMLAttribType.STR: str,
    XMLAttribType.BOOL: to_bool,
    XMLAttribType.INT: int,
    XMLAttribType.FLOAT: float,
    XMLAttribType.STR_LIST  : lambda s: [i.strip() for i in s.split(",") if i],
    XMLAttribType.INT_LIST  : lambda s: [int(i) for i in s.split(",") if i],
    XMLAttribType.FLOAT_LIST: lambda s: [float(i) for i in s.split(",") if i]
}

class XMLWSBase(AbstractObject):
    
    EPSILON = 1e-6
    
    def __init__(self, source:Optional[Union[str, Dict]]=None, basedir:Optional[str]=None,
                 unlimited_stack:bool=True, verbosity:Optional[Union[int, str]]="INFO"):
        super().__init__(verbosity=verbosity)
        self.source = source
        if basedir is not None:
            self.basedir = basedir
        # source is a filepath
        elif isinstance(source, str):
            self.basedir = os.path.dirname(source)
        else:
            self.basedir = os.getcwd()
        self.path_manager = PathManager(self.basedir)
        if unlimited_stack:
            set_unlimited_stacksize()
            self.stdout.info("INFO: Set stack size to unlimited")
        self.suppress_roofit_message()
        self.custom_classes = {}

    def load_extension(self):
        extensions = quickstats.get_workspace_extensions()
        for extension in extensions:
            result = load_macro(extension)
            if (result is not None) and hasattr(ROOT, extension):
                self.stdout.info(f'INFO: Loaded extension module "{extension}"')
            try:
                cls = get_macro_TClass(extension)
            except:
                continue
            self.custom_classes[extension] = cls
    
    @staticmethod
    def suppress_roofit_message():
        ROOT.RooMsgService.instance().getStream(1).removeTopic(ROOT.RooFit.NumIntegration)
        ROOT.RooMsgService.instance().getStream(1).removeTopic(ROOT.RooFit.Fitting)
        ROOT.RooMsgService.instance().getStream(1).removeTopic(ROOT.RooFit.Minimization)
        ROOT.RooMsgService.instance().getStream(1).removeTopic(ROOT.RooFit.InputArguments)
        ROOT.RooMsgService.instance().getStream(1).removeTopic(ROOT.RooFit.Eval)
        ROOT.RooMsgService.instance().setGlobalKillBelow(ROOT.RooFit.ERROR)
        
    def get_relpath(self, filename:str):
        if filename.startswith("/") or filename.startswith("~"):
            return filename
        return os.path.join(self.basedir, filename)
    
    def check_file_exist(self, filename:str):
        return os.path.exists(self.get_relpath(filename))
    
    @staticmethod
    def _get_node_attrib(node:Dict, attrib_name:Union[str, List[str]], required:bool=True,
                         default:Optional[str]=None,
                         dtype:Union[XMLAttribType,str]='str'):
        attrib_node = node['attrib']
        if isinstance(attrib_name, str):
            attrib_name = [attrib_name]
        # aliases
        found_attribute = False
        for name in attrib_name:
            if name in attrib_node:
                if found_attribute:
                    raise RuntimeError("values defined for multiple alises of the same attribute: "
                                      f"{','.join(attrib_name)}")
                attrib_val = attrib_node[name]
                found_attribute = True
        if not found_attribute:
            if required:
                raise RuntimeError(f"attribute `{','.join(attrib_name)}` not found in the "
                                   f"node `{node['tag']}`")
            else:
                attrib_val = default
        if not isinstance(attrib_val, str):
            return attrib_val
        attrib_val = re.sub('\s+', '', attrib_val)
        dtype = XMLAttribType.parse(dtype)
        attrib_val = ConversionOperators[dtype](attrib_val)
        return attrib_val
    
    @staticmethod
    def _get_object_name_and_type_from_expr(expr:str):
        if ("::" in expr) and ("(" in expr):
            object_type = WSObjectType.FUNCTION
            object_name = expr.split("::")[1].split("(")[0]
        elif ("(" in expr):
            object_type = WSObjectType.CONSTRAINT
            object_name = expr.split("(")[0]
        elif ("[" in expr):
            object_type = WSObjectType.VARIABLE
            object_name = expr.split("[")[0]
        elif (":" in expr) and not ("::" in expr):
            raise RuntimeError(f"syntax error for the expression `{expr}`: missing colon pair")
        else:
            object_type = WSObjectType.DEFINED
            object_name = expr
        return object_name, object_type
    
    @staticmethod
    def _extract_bracket_expr(expr:str, bracket_type:Union[str, BracketType]=BracketType.ROUND,
                              match:bool=True):
        bracket_type = BracketType.parse(bracket_type)
        if match:
            result = bracket_type.regex.match(expr)
        else:
            result = bracket_type.regex.search(expr)
        if not result:
            return (None, None)
        else:
            return (result.group(1), result.group(2))
    
    def import_object(self, ws:ROOT.RooWorkspace, obj, *args, silent:bool=True):
        if silent:
            getattr(ws, "import")(obj, *args, ROOT.RooFit.Silence())        
        else:
            getattr(ws, "import")(obj, *args)    
    
    def import_expression(self, ws:ROOT.RooWorkspace, object_expr:str, check_defined:bool=False):
        object_name, object_type = self._get_object_name_and_type_from_expr(object_expr)
        # throw error when attemping to access undefined object
        if object_type == "defined":
            if (not ws.arg(object_name)):
                raise RuntimeError(f"object `{object_name}` does not exist")
            else:
                return object_name
        # throw error? when attemping to define already existing object
        if check_defined:
            if ws.arg(object_name):
                self.stdout.debug(f"REGTEST: object {object_name} already exists")
                return object_name
        self.stdout.debug(f"REGTEST: Generating {object_type} {object_expr}")
        status = ws.factory(object_expr)
        if not status:
            raise RuntimeError(f"object creation from expression `{object_expr}` had failed")
        return object_name
    
    @semistaticmethod
    def get_name_mapping_str(self, old_name:str, new_name:str, width:int=40):
        sequence = difflib.SequenceMatcher(None, old_name, new_name)
        match = sequence.find_longest_match(0, len(old_name), 0, len(new_name))
        ahead = old_name[ : match.a]
        atail = old_name[match.a + match.size : ]
        bhead = new_name[ : match.b]
        btail = new_name[match.b + match.size : ]
        common1 = old_name[match.a : match.a+match.size]
        common2 = new_name[match.b : match.b+match.size]
        result = f'{ahead}\033[91m{common1}\033[0m{atail}'.rjust(width, ' ') + ' -> ' + f'{bhead}\033[92m{common1}\033[0m{btail}'
        return result
    
    @semistaticmethod
    def get_name_mapping_str_arrays(self, rename_map:Dict):
        results = []
        if not rename_map:
            return results        
        width = max([len(i) for i in list(rename_map)]) + 20
        for old_name, new_name in rename_map.items():
            results.append(self.get_name_mapping_str(old_name, new_name, width))
        return results
    
    def get_workspace_arg(self, ws:ROOT.RooWorkspace, arg_name:str, strict:bool=True):
        obs = ws.arg(arg_name)
        if not arg_name:
            if strict:
                raise RuntimeError(f"variable {arg_name} does not exist in workspace")
            return None
        return obs
    
    def import_expressions(self, ws:ROOT.RooWorkspace, object_expr_list:List[str],
                           check_defined:bool=False):
        object_names = []
        for object_expr in object_expr_list:
            object_name = self.import_expression(ws, object_expr, check_defined=check_defined)
            object_names.append(object_name)
        return ",".join(object_names)
    
    def import_class_code(self, ws:ROOT.RooWorkspace):
        all_function_class = [i.ClassName().split("::")[-1] for i in ws.allFunctions()]
        all_pdf_class = [i.ClassName().split("::")[-1] for i in ws.allPdfs()]
        all_class = all_function_class + all_pdf_class
        for extension, cls in self.custom_classes.items():
            # the workspace does not use the extended class, no need to import
            if extension not in all_class:
                continue
            macro_dir = get_macro_dir(extension)
            ws.addClassDeclImportDir(macro_dir)
            ws.addClassImplImportDir(macro_dir)
            ws.importClassCode(cls)
            self.stdout.info(f"INFO: Imported class code for \"{cls.GetName()}\".")
        #ws.importClassCode()    
    
    def _generate_asimov(self, ws:ROOT.RooWorkspace, asimov_definitions:List, data_name:str,
                         range_name:Optional[str]=None, minimizer_config:Optional[Dict]=None,
                         title_indent_str:str="\t"):
        from quickstats.components.workspaces import AsimovHandler
        n_asimov = len(asimov_definitions)
        if n_asimov == 0:
            return None
        asimov_handler = AsimovHandler(ws, data_name=data_name, range_name=range_name,
                                       minimizer_config=minimizer_config,
                                       verbosity=self.stdout.verbosity)
        asimov_handler.title_indent_str = title_indent_str
        for attributes in asimov_definitions:
            asimov_handler.generate_single_asimov(attributes)


from typing import Optional, List, Dict

from .rooproc_base_action import RooProcBaseAction
from .auxiliary import RooProcReturnCode

class RooProcNestedAction(RooProcBaseAction):
    
    def __init__(self, **params):
        super().__init__(**params)

    def _execute(self, processor:"quickstats.RooProcessor", **params):
        return RooProcReturnCode.NORMAL
    
    def execute(self, processor:"quickstats.RooProcessor", global_vars:Optional[Dict]=None):
        params = self.get_formatted_parameters(global_vars)
        return_code = self._execute(processor, **params)
        self.executed = True
        return return_code


from typing import Optional
import re

whitespace_trans = str.maketrans('', '', " \t\r\n\v")

def split_lines(string:str, comment_string:Optional[str]="#", remove_blank:bool=True,
                with_line_number:bool=False):
    lines = string.splitlines(True)
    # remove comments
    lines = [l.split(comment_string)[0] for l in lines]
    # remove blank lines
    if with_line_number:
        if remove_blank:
            return [(l, i+1) for i, l in enumerate(lines) if not re.match(r'^\s*$', l)]
        return [(l, i+1) for i, l in enumerate(lines)]
    if remove_blank:
        lines = filter(lambda x: not re.match(r'^\s*$', x), lines)
    return lines


def split_str(string:str, strip:bool=True, remove_empty:bool=False):
    if strip:
        tokens = [s.strip() for s in string.strip().split()]
    else:
        tokens = string.split()
    if remove_empty:
        # remove empty tokens
        tokens = [s for s in tokens if s]
    return tokens

def remove_whitespace(string:str):
    return string.translate(whitespace_trans)


from typing import Optional

from .rooproc_helper_action import RooProcHelperAction

class RooProcSaveFrame(RooProcHelperAction):
    def __init__(self, name:str):
        super().__init__(name=name)

    @classmethod
    def parse(cls, main_text:str, block_text:Optional[str]=None):
        return cls(name=main_text)
    
    def _execute(self, processor:"quickstats.RooProcessor", **params):
        frame_name = params['name']
        if frame_name in processor.rdf_frames:
            processor.stdout.warning(f"WARNING: Overriding existing rdf frame `{frame_name}`")
        processor.rdf_frames[frame_name] = processor.rdf
        return processor


from typing import Optional, Union, List, Dict

import ROOT

from quickstats.interface.cppyy.vectorize import list2vec

from .systematic import Systematic
from .settings import RESPONSE_PREFIX, GLOBALOBS_PREFIX, CONSTRTERM_PREFIX

# basic data structure for systematic domain
# all the systematics under the same systematic domain will be merged into one response function          
class SystematicsDomain:
    
    def __init__(self, domain:str):
        self.domain = domain
        self.is_shape = None
        self.nuis_list = ROOT.RooArgList()
        self.nominal_list = []
        self.uncert_hi_list = ROOT.RooArgList()
        self.uncert_lo_list = ROOT.RooArgList()
        self.interp_code_list = []
        self.constr_term_list = []
        
    def get_response_name(self):
        return f"{RESPONSE_PREFIX}{self.domain}"
        
    def get_response_function(self):
        if not getattr(ROOT, "ResponseFunction"):
            raise RuntimeError("ResponseFunction class undefined: may be you need to load the corresponding "
                               "macro...")
        nominal_list = list2vec(self.nominal_list)
        interp_code_list = list2vec(self.interp_code_list)
        resp_name = self.get_response_name()
        resp_func = ROOT.ResponseFunction(resp_name, resp_name, self.nuis_list, nominal_list,
                                          self.uncert_lo_list, self.uncert_hi_list,
                                          interp_code_list)
        return resp_func
    
    def get_np_glob_constr_names(self, idx:int):
        if idx >= self.nuis_list.size():
            raise RuntimeError("index out of range")
        np = self.nuis_list.at(idx)
        np_name = np.GetName()
        glob_name = f"{GLOBALOBS_PREFIX}{np_name}"
        constr_name = f"{CONSTRTERM_PREFIX}{np_name}"
        return np_name, glob_name, constr_name
    
    def add_item(self, nuis:ROOT.RooRealVar, nominal:float, uncert_hi:ROOT.RooRealVar,
                 uncert_lo:ROOT.RooRealVar, interp_code:int, constr_term:str):
        self.nuis_list.add(nuis)
        self.nominal_list.append(nominal)
        self.uncert_hi_list.add(uncert_hi)
        self.uncert_lo_list.add(uncert_lo)
        self.interp_code_list.append(interp_code)
        self.constr_term_list.append(constr_term)


from typing import Optional, Union, Dict, List

import matplotlib.patches as patches
import matplotlib.lines as lines
import numpy as np
import pandas as pd

from quickstats.plots.template import create_transform
from quickstats.plots import AbstractPlot
from quickstats.utils.common_utils import combine_dict


class UpperLimit1DPlot(AbstractPlot):

    STYLES = {
        'figure': {
            'figsize': (11.111, 10.333),
            'dpi': 72,
            'facecolor': "#FFFFFF"
        },
        'axis': {
            'tick_bothsides': False
        },
        'legend': {
            'fontsize': 22
        },
        'text': {
            'fontsize': 22
        }
    }

    COLOR_PALLETE = {
        '2sigma': 'hh:darkyellow',
        '1sigma': 'hh:lightturquoise',
        'expected': 'k',
        'third': 'k',
        'observed': 'k',
    }

    LABELS = {
        '2sigma': r'Expected limit $\pm 2\sigma$',
        '1sigma': r'Expected limit $\pm 1\sigma$',
        'expected': r'Expected limit',
        'third': 'Third limit',
        'observed': r'Observed limit',
    }

    CONFIG = {
        'top_margin': 2.2,
        'curve_line_styles': {
            'color': 'darkred'
        },
        'curve_fill_styles': {
            'color': 'hh:darkpink'
        },
    }

    def __init__(self, category_df, label_map, line_below=None,
                 color_pallete: Optional[Dict] = None,
                 labels: Optional[Dict] = None,
                 config: Optional[Dict] = None,
                 styles: Optional[Union[Dict, str]] = None,
                 analysis_label_options: Optional[Union[Dict, str]] = None):
        super().__init__(color_pallete=color_pallete,
                         styles=styles,
                         analysis_label_options=analysis_label_options)
        self.category_df = category_df
        self.label_map = label_map
        self.line_below = line_below
        self.curve_data = None

        self.labels = combine_dict(self.LABELS, labels)
        self.config = combine_dict(self.CONFIG, config)

    def get_default_legend_order(self):
        return ["observed", "expected", "third", "one_sigma", "two_sigma", "curve"]        

    def add_curve(self, x, xerrlo=None, xerrhi=None,
                  label: str = "Theory prediction",
                  line_styles: Optional[Dict] = None,
                  fill_styles: Optional[Dict] = None):
        curve_data = {
            'x': x,
            'y': np.arange(0, len(self.category_df.columns)+1),
            'xerrlo': xerrlo,
            'xerrhi': xerrhi,
            'label': label,
            'line_styles': line_styles,
            'fill_styles': fill_styles,
        }
        self.curve_data = curve_data

    def draw_curve(self, ax, data):
        line_styles = data['line_styles']
        fill_styles = data['fill_styles']
        if line_styles is None:
            line_styles = self.config['curve_line_styles']
        if fill_styles is None:
            fill_styles = self.config['curve_fill_styles']
        if (data['xerrlo'] is None) and (data['xerrhi'] is None):
            line_styles['color'] = fill_styles['color']
        handle_line = ax.vlines(
            data['x'], data['y'][0], data['y'][-1], label=data['label'], **line_styles)
        handles = handle_line
        if (data['xerrlo'] is not None) and (data['xerrhi'] is not None):
            handle_fill = ax.fill_betweenx(data['y'], data['xerrlo'], data['xerrhi'],
                                           label=data['label'], **fill_styles)
            handles = (handle_fill, handle_line)
        self.update_legend_handles({'curve': handles})

    def draw(self, logx:bool=False, xlabel:Optional[str]=None, markersize:float=50.,
             draw_observed:bool=True, draw_stat:bool=False, draw_third_column:Optional[str]=None,
             add_text:bool= True, sig_fig:int=2):
        if (draw_observed + draw_stat) > 1:
            raise RuntimeError(
                "draw_observed and draw_stat can not be both True")
        n_category = len(self.category_df.columns)
        ax = self.draw_frame(logx=logx)
        transform = create_transform(transform_x='axis', transform_y='data')

        if draw_observed:
            text_pos = {'observed': 0.775, 'expected': 0.925}
        if draw_stat:
            text_pos = {'expected': 0.775, 'stat': 0.925}
        if (not draw_observed) and (not draw_stat):
            text_pos = {'expected': 0.925}
        if draw_third_column:
            text_pos = {'observed': 0.725, 'expected': 0.825, 'third': 0.925}

        for i, category in enumerate(self.category_df):
            df = self.category_df[category]
            # draw observed
            if draw_observed:
                observed_limit = df['obs']
                handle_1 = ax.vlines(observed_limit, i, i+1, colors=self.color_pallete['observed'], linestyles='solid',
                                     zorder=1.1, label=self.labels['observed'] if i == 0 else '')
                handle_2 = ax.scatter(observed_limit, i + 0.5, s=markersize, marker='o',
                                      color=self.color_pallete['observed'], zorder=1.1)
                observed_handle = (handle_1, handle_2)
                if add_text:
                    ax.text(text_pos['observed'], i + 0.5, f"{{:.{sig_fig}f}}".format(observed_limit),
                            horizontalalignment='center',
                            verticalalignment='center',
                            transform=transform,
                            **self.styles['text'])
            else:
                observed_handle = None
            # draw stat
            if draw_stat:
                stat_limit = df['stat']
                if add_text:
                    ax.text(text_pos['stat'], i + 0.5, f"({{:.{sig_fig}f}})".format(stat_limit),
                            horizontalalignment='center',
                            verticalalignment='center',
                            transform=transform,
                            **self.styles['text'])
            # draw expected
            expected_limit = df['0']
            expected_handle = ax.vlines(expected_limit, i, i + 1, colors=self.color_pallete['expected'], linestyles='dotted',
                                        zorder=1.1, label=self.labels['expected'])
            if add_text:
                ax.text(text_pos['expected'], i + 0.5, f"{{:.{sig_fig}f}}".format(expected_limit),
                        horizontalalignment='center',
                        verticalalignment='center',
                        transform=transform,
                        **self.styles['text'])
            # draw third
            if draw_third_column:
                third_limit = df['third']
                third_handle = ax.vlines(third_limit, i, i + 1, colors=self.color_pallete['third'], linestyles='dashed',
                                         zorder=1.1, label=self.labels['third'])
                if add_text:
                    ax.text(text_pos['third'], i + 0.5, f"{{:.{sig_fig}f}}".format(third_limit),
                            horizontalalignment='center',
                            verticalalignment='center',
                            transform=transform,
                            **self.styles['text'])
            else:
                third_handle = None
            # draw error band
            two_sigma_handle = ax.fill_betweenx([i, i + 1], df['-2'], df['2'], facecolor=self.color_pallete['2sigma'],
                                                label=self.labels['2sigma'])
            
            one_sigma_handle = ax.fill_betweenx([i, i + 1], df['-1'], df['1'], facecolor=self.color_pallete['1sigma'],
                                                label=self.labels['1sigma'])
            if i == 0:
                handles = {
                    "expected": expected_handle,
                    "one_sigma": one_sigma_handle,
                    "two_sigma": two_sigma_handle,
                }
                for key, handle in [("observed", observed_handle), ("third", third_handle)]:
                    if handle is not None:
                        handles[key] = handle
                self.update_legend_handles(handles)
        xlim = ax.get_xlim()
        ax.set_xlim(xlim[0] - (xlim[1]/0.7 - xlim[1])*0.5, xlim[1]/0.7)
        ax.set_ylim(0, len(self.category_df.columns) +
                    self.config['top_margin'])
        ax.set_yticks(np.arange(n_category) + 0.5, minor=False)
        ax.tick_params(axis="y", which="minor", length=0)
        for axis in ['top', 'bottom', 'left', 'right']:
            ax.spines[axis].set_linewidth(2)
        ax.set_yticklabels([self.label_map[i] for i in self.category_df.columns.to_list()],
                           horizontalalignment='right')
        # draw horizonal dashed lines
        ax.axhline(n_category, color='k', ls='--', lw=1)
        if self.line_below is not None:
            for category in self.line_below:
                position = np.where(
                    np.array(self.category_df.columns, dtype='str') == category)[0]
                if position.shape[0] != 1:
                    raise ValueError(
                        "category `{}` not found in dataframe".format(category))
                ax.axhline(position[0], color='k', ls='--', lw=1)
        if add_text:
            if draw_observed:
                ax.text(text_pos['observed'], n_category + 0.3, 'Obs.',
                        horizontalalignment='center',
                        verticalalignment='center',
                        transform=transform,
                        **self.styles['text'])
            if draw_stat:
                ax.text(text_pos['stat'], n_category + 0.3, '(Stat.)',
                        horizontalalignment='center',
                        verticalalignment='center',
                        transform=transform,
                        **self.styles['text'])
            if draw_third_column:
                ax.text(text_pos['third'], n_category + 0.3, draw_third_column,
                        horizontalalignment='center',
                        verticalalignment='center',
                        transform=transform,
                        **self.styles['text'])
            ax.text(text_pos['expected'], n_category + 0.3, 'Exp.',
                    horizontalalignment='center',
                    verticalalignment='center',
                    transform=transform,
                    **self.styles['text'])
        if self.curve_data is not None:
            self.draw_curve(ax, self.curve_data)
        if xlabel is not None:
            ax.set_xlabel(xlabel, **self.styles['xlabel'])
        # border for the legend
        border_leg = patches.Rectangle( (0, 0), 1, 1, facecolor='none', edgecolor='black', linewidth=1)
        self.add_legend_decoration(border_leg, targets=["one_sigma", "two_sigma", "curve"])
        handles, labels = self.get_legend_handles_labels()
        ax.legend(handles, labels, **self.styles['legend'])
        return ax



from typing import Optional, List, Dict, Union
import os
import glob
import json
import time
import ROOT

from .builtin_methods import BUILTIN_METHODS
from .actions import *
from .roo_config_parser import RooConfigParser

from quickstats import Timer, AbstractObject
from quickstats.utils.root_utils import declare_expression, close_all_root_files

class RooProcessor(AbstractObject):
    def __init__(self, config_path:Optional[str]=None,
                 flags:Optional[List[str]]=None,
                 multithread:bool=True,
                 cache:bool=False,
                 use_template:bool=False,
                 verbosity:Optional[Union[int, str]]="INFO"):
        super().__init__(verbosity=verbosity)
        self.cache = cache
        self.action_tree = None
        if flags is not None:
            self.flags = list(flags)
        else:
            self.flags = []
        self.rdf_frames = {}
        self.rdf = None
        self.global_variables = {}
        self.external_variables = {}
        self.treename = None
        self.use_template = use_template
        
        self.load_buildin_functions()
        
        if multithread:
            ROOT.EnableImplicitMT()
        
        if config_path is not None:
            self.load_config(config_path)
            
    def set_cache(self, cache:bool=True):
        self.cache = cache
            
    def load_buildin_functions(self):
        for name, definition in BUILTIN_METHODS.items():
            declare_expression(definition, name)
    
    def load_config(self, config_path:Optional[str]=None):
        action_tree = RooConfigParser.parse_file(config_path)
        action_tree.construct_actions()
        if not action_tree.root_node.has_child:
            raise RuntimeError("no actions found in the process card")
        first_action = action_tree.root_node.first_child.action
        if not isinstance(first_action, RooProcTreeName):
            raise RuntimeError("tree name must be specified at the beginning of the process card")
        self.treename = first_action._params['treename']
        self.action_tree = action_tree
        
    def set_global_variables(self, **kwargs):
        self.global_variables.update(kwargs)
        
    def clear_global_variables(self):
        self.global_variables = {}
    
    def add_flags(self, flags:List[str]):
        self.flags += list(flags)
        
    def set_flags(self, flags:List[str]):
        self.flags = list(flags)        
        
    def cleanup(self):
        close_all_root_files()
        self.rdf_frames = {}
        self.rdf = None
        
    def _get_all_files(self, filename:Union[List[str], str]):
        all_files = []
        if isinstance(filename, str):
            filename = [filename]
        if not isinstance(filename, list):
            raise ValueError("filename must be either string or list of strings")
        for fname in filename:
            if os.path.isdir(fname):
                all_files += glob.glob(os.path.join(fname, "*.root"))
            else:
                all_files += glob.glob(fname)
        if not all_files:
            raise FileNotFoundError(f'file "{filename}" does not exist')
        return all_files
    
    def run_action(self, action:RooProcBaseAction):
        if not self.rdf:
            raise RuntimeError("RDataFrame instance not initialized")
        if isinstance(action, RooProcRDFAction):
            self.rdf = action.execute(self.rdf, self.global_variables)
        elif isinstance(action, RooProcHelperAction):
            action.execute(self, self.global_variables)
        elif isinstance(action, RooProcHybridAction):
            self.rdf, _ = action.execute(self.rdf, self, self.global_variables)
        elif isinstance(action, RooProcNestedAction):
            return_code = action.execute(self, self.global_variables)
            return return_code
        else:
            raise RuntimeError("unknown action type")
        return RooProcReturnCode.NORMAL
            
    def run_all_actions(self, consider_child:bool=True):
        if not self.action_tree:
            raise RuntimeError("action tree not initialized")
        node = self.action_tree.get_next(consider_child=consider_child)
        if node is not None:
            self.stdout.debug(f'DEBUG: Executing node "{node.name}" defined at line {node.data["start_line_number"]}')
            action = node.action
            return_code = self.run_action(action)
            if return_code == RooProcReturnCode.NORMAL:
                self.run_all_actions()
            elif return_code == RooProcReturnCode.SKIP_CHILD:
                self.run_all_actions(consider_child=False)
            else:
                raise RuntimeError("unknown return code")
        else:
            self.stdout.debug(f'DEBUG: All node executed')
            
    def sanity_check(self):
        if not self.action_tree:
            raise RuntimeError("action tree not initialized")        
        if not self.action_tree.root_node.has_child:
            self.stdout.warning("WARNING: No actions to be performed.")
            return None
        if self.treename is None:
            raise RuntimeError("tree name is undefined")
    
    def run(self, filename:Union[List[str], str]):
        self.sanity_check()
        all_files = self._get_all_files(filename)
        if len(all_files) == 1:
            self.stdout.info(f'INFO: Processing file "{all_files[0]}".')
        else:
            self.stdout.info(f"INFO: Professing files")
            for f in all_files:
                self.stdout.info(f'\t"{f}"')  
        with Timer() as t:
            self.rdf = ROOT.RDataFrame(self.treename, all_files)
            self.action_tree.reset()
            self.run_all_actions()
            self.cleanup()
            
        self.stdout.info(f"INFO: Task finished. Total time taken: {t.interval:.3f} s.")


from typing import Optional

import pandas as pd

from .rooproc_hybrid_action import RooProcHybridAction

from quickstats.utils.common_utils import is_valid_file

class RooProcReport(RooProcHybridAction):
    def __init__(self, display:bool=False, filename:Optional[str]=None):
        super().__init__(display=display,
                         filename=filename)
    @classmethod
    def parse(cls, main_text:str, block_text:Optional[str]=None):
        kwargs = cls.parse_as_kwargs(main_text)
        return cls(**kwargs)
    
    def _execute(self, rdf:"ROOT.RDataFrame", processor:"quickstats.RooProcessor", **params):
        display = params['display']
        filename = params['filename']
        if processor.cache and is_valid_file(filename):
            processor.stdout.info(f"INFO: Cached output `{filename}`.")
            return rdf, processor        
        cut_report = rdf.Report()
        result = []
        cumulative_eff  = 1
        for report in cut_report:
            data = {}
            data['name'] = report.GetName()
            data['all']  = report.GetAll()
            data['pass'] = report.GetPass()
            data['efficiency'] = report.GetEff()
            cumulative_eff *= data['efficiency']/100
            data['cumulative_efficiency'] = cumulative_eff*100
            result.append(data)
        df = pd.DataFrame(result)
        if int(display):
            processor.stdout.info(df)
        if filename is not None:
            self.makedirs(filename)
            processor.stdout.info(f'INFO: Writing cutflow report to "{filename}".')
            df.to_csv(filename, index=False)
        return rdf, processor


from typing import Dict, Optional, Union, List, Tuple
import pandas as pd
import numpy as np

from quickstats.plots import AbstractPlot, StatPlotConfig
from quickstats.plots.template import create_transform, handle_has_label
from quickstats.utils.common_utils import combine_dict

class General1DPlot(AbstractPlot):
    
    def __init__(self, data_map:Union[pd.DataFrame, Dict[str, pd.DataFrame]],
                 label_map:Optional[Dict]=None,
                 styles_map:Optional[Dict]=None,
                 color_cycle=None,
                 styles:Optional[Union[Dict, str]]=None,
                 analysis_label_options:Optional[Dict]=None,
                 config:Optional[Dict]=None):
        
        self.data_map = data_map
        self.label_map = label_map
        self.styles_map = styles_map
        
        super().__init__(color_cycle=color_cycle,
                         styles=styles,
                         analysis_label_options=analysis_label_options,
                         config=config)
        
        self.stat_configs = {}
        
    def get_default_legend_order(self):
        if not isinstance(self.data_map, dict):
            return []
        else:
            return list(self.data_map)
        
    def configure_stats(self, stat_configs:List[StatPlotConfig],
                        targets:Optional[Union[str, List[str]]]=None,
                        extend:bool=True):
        if not isinstance(targets, list):
            targets = [targets]
        for target in targets:
            if extend and (target in self.stat_configs):
                self.stat_configs[target].extend(stat_configs)
            else:
                self.stat_configs[target] = stat_configs
        
    def draw_single_data(self, ax, data:pd.DataFrame,
                         xattrib:str, yattrib:str,
                         stat_configs:Optional[List[StatPlotConfig]]=None,
                         styles:Optional[Dict]=None,
                         label:Optional[str]=None):
        x = data[xattrib].values
        y = data[yattrib].values
        indices = np.argsort(x)
        x = x[indices]
        y = y[indices]
        draw_styles = combine_dict(self.styles['plot'], styles)
        handle = ax.plot(x, y, **draw_styles, label=label)
        if stat_configs is not None:
            stat_handles = []
            for stat_config in stat_configs:
                stat_config.set_data(y)
                stat_handle = stat_config.apply(ax, handle[0])
                stat_handles.append(stat_handle)
        else:
            stat_handles = None
        return handle[0], stat_handles
    
    def draw(self, xattrib:str, yattrib:str, targets:Optional[List[str]]=None,
             xlabel:Optional[str]=None, ylabel:Optional[str]=None,
             ymin:Optional[float]=None, ymax:Optional[float]=None,
             xmin:Optional[float]=None, xmax:Optional[float]=None,
             draw_stats:bool=True):
        
        ax = self.draw_frame()
        
        legend_order = []
        if isinstance(self.data_map, pd.DataFrame):
            if draw_stats and (None in self.stat_configs):
                stat_configs = self.stat_configs[None]
            else:
                stat_configs = None
            handle, stat_handles = self.draw_single_data(ax, self.data_map, xattrib=xattrib, yattrib=yattrib,
                                                         stat_configs=stat_configs,
                                                         styles=self.styles_map)
        elif isinstance(self.data_map, dict):
            if targets is None:
                targets = list(self.data_map.keys())
            if self.styles_map is None:
                styles_map = {k:None for k in self.data_map}
            else:
                styles_map = self.styles_map
            if self.label_map is None:
                label_map = {k:k for k in self.data_map}
            else:
                label_map = self.label_map
            handles = {}
            for target in targets:
                data = self.data_map[target]
                styles = styles_map.get(target, None)
                label = label_map.get(target, "")
                if draw_stats:
                    if target in self.stat_configs:
                        stat_configs = self.stat_configs[target]
                    elif None in self.stat_configs:
                        stat_configs = self.stat_configs[None]
                    else:
                        stat_configs = None
                else:
                    stat_configs = None
                handle, stat_handles = self.draw_single_data(ax, data, 
                                                             xattrib=xattrib,
                                                             yattrib=yattrib,
                                                             stat_configs=stat_configs,
                                                             styles=styles,
                                                             label=label)
                handles[target] = handle
                if stat_handles is not None:
                    for i, stat_handle in enumerate(stat_handles):
                        if handle_has_label(stat_handle):
                            handle_name = f"{target}_stat_handle_{i}"
                            handles[handle_name] = stat_handle
            legend_order.extend(handles.keys())
            self.update_legend_handles(handles)
        else:
            raise ValueError("invalid data format")
            
        self.legend_order = legend_order
        handles, labels = self.get_legend_handles_labels()
        ax.legend(handles, labels, **self.styles['legend'])
        
        self.draw_axis_components(ax, xlabel=xlabel, ylabel=ylabel)
        self.set_axis_range(ax, xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax)
        
        return ax



from typing import Optional, Union, List, Dict

from quickstats import DescriptiveEnum
from quickstats.maths.numerics import is_float

from .settings import (NORM_SYST_KEYWORD, SHAPE_SYST_KEYWORD, COMMON_SYST_DOMAIN_NAME,
                       CONSTR_GAUSSIAN, CONSTR_LOGN, CONSTR_ASYM, CONSTR_DFD, OTHER,
                       UNCERT_HI_PREFIX, UNCERT_LO_PREFIX, UNCERT_SYM_PREFIX)

class SystematicType(DescriptiveEnum):
    Norm  = (0, "Normalization sysetmatic", NORM_SYST_KEYWORD)
    Shape = (1, "Shape systematic", SHAPE_SYST_KEYWORD)
    Other = (2, "Unclassified systematic type", OTHER)
    
    def __new__(cls, value:int, description:str, keyword:str):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.description = description
        obj.keyword = keyword
        return obj    
    
class ConstraintType(DescriptiveEnum):
    LogN  = (0, "Lognormal constraint", CONSTR_LOGN, 1)
    Asym  = (1, "Assymetric constraint", CONSTR_ASYM, 4)
    Gaus  = (2, "Gaussian constraint", CONSTR_GAUSSIAN, 0)
    DFD   = (3, "Double Fermi-Dirac", CONSTR_DFD, 0)
    Other = (4, "Unclassified constraint type", OTHER, -1)
    
    def __new__(cls, value:int, description:str, keyword:str, interp_code:int):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.description = description
        obj.keyword = keyword
        obj.interp_code = interp_code
        return obj

class Systematic:
    
    def __init__(self):
        self.name = ""
        self.domain = ""
        self.raw_domain = ""
        self.process = ""
        self.whereto = ""
        self.constr_term = ""
        self.nominal = 0.
        self.beta = 0.
        self.errorlo = None
        self.errorhi = None
    
    def __eq__(self, other:"Systematic"):
        return (self.name == other.name) and (self.process == other.process) and (self.whereto == other.whereto)
    
    @staticmethod
    def default_domain():
        return COMMON_SYST_DOMAIN_NAME
    
    @staticmethod
    def common_domain():
        return COMMON_SYST_DOMAIN_NAME
    
    def is_equal(self, other:"Systematic"):
        return self.__eq__(other)
    
    def is_shape_syst(self):
        return self.get_syst_type() == SystematicType.Shape

    def get_syst_type(self):
        result = SystematicType.get_member_by_attribute("keyword", self.whereto)
        if result is None:
            return SystematicType.Other
        return result
    
    def get_constr_type(self):
        result = ConstraintType.get_member_by_attribute("keyword", self.constr_term)
        if result is None:
            return ConstraintType.Other
        return result
    
    def is_common_domain(self):
        return self.domain == COMMON_SYST_DOMAIN_NAME
    
    def set_domain(self, domain:str):
        syst_type = self.get_syst_type()
        # systematics defined in the common domain
        if domain == self.common_domain():
            # if a process name is specified, use it as domain name and 
            # remove it from common systematic
            if self.process != "":
                if syst_type == SystematicType.Shape:
                    self.domain = f"{self.whereto}_{self.name}_{self.process}"
                else:
                    self.domain = self.process
            # otherwise consider it as common systematic
            else:
                # for shape uncertainty, use NP name as domain
                if syst_type == SystematicType.Shape:
                    self.domain = f"{self.whereto}_{self.name}"
                # for yield uncertainty, consider it as common systematic
                elif syst_type == SystematicType.Norm:
                    self.domain = domain
                else:
                    raise RuntimeError(f'Unknown systematic loacation {self.whereto} '
                                       f'for NP {self.name}. Choose from "{NORM_SYST_KEYWORD}" (NormSyst) '
                                       f'or "{SHAPE_SYST_KEYWORD}" (ShapeSyst)')
        else:
            # for yield systematics under a sample, the domain name is always the sample name
            # for shape systematics under a sample, we will update it later depending on whether "Process" is specified
            self.domain = domain
            # for yield systematics under a sample, the default process name is the sample name
            # if a process name is provided in addition (e.g. uncertainties on acceptance and correction factor are separately provided in differential xs measurements), it will be <sample_name>_<process_name>
            if self.process == "":
                self.process = domain
                if syst_type == SystematicType.Shape:
                    self.domain = f"{self.whereto}_{self.name}_{self.process}"
            else:
                self.process = f"{domain}_{self.process}"
                if syst_type == SystematicType.Shape:
                    self.domain = f"{self.whereto}_{self.name}_{self.process}"
        self.raw_domain = domain
        
    def set_magnitudes(self, values:Union[float, str, List[float], List[str]]):
        # for asymmetric uncertainty, we need to be careful with the relative sign between the two components
        # if both upper and lower uncertainties are numbers, the lower number will always be forced to have opposite sign of upper
        # if either uncertainty is implemented as formula, the neither component can bear sign. Users need to make sure the sign convention is consistent
        if not isinstance(values, list):
            values = [values]
        if len(values) == 1:
            value = values[0]
            if isinstance(value, str):
                if is_float(value):
                    self.errorhi = float(value)
                    self.errorlo = float(value)
                else:
                    self.errorhi = value
                    self.errorlo = value
            elif isinstance(value, float):
                self.errorhi = value
                self.errorlo = value
            else:
                raise ValueError(f"invalid systematic magnitude: {value}")
        elif len(values) == 2:
            errhi, errlo = values
            # both are numbers, need to handle the relative sign
            if is_float(errhi) and is_float(errlo):
                if float(errhi) < 0:
                    self.errorhi = float(errhi)
                    self.errorlo = abs(float(errlo))
                else:
                    self.errorhi = float(errhi)
                    self.errorlo = -abs(float(errlo))
            else:
                self.errorhi = errhi
                self.errorlo = errlo
            if self.get_constr_type() != ConstraintType.Asym:
                self.stdout.warn(f"WARNING: The systematic {self.name} under domain {self.domain} "
                                 f"will be implemented as an asymmetric uncertainty for now. Please "
                                 f"double-check your config file and use the keyword "
                                 f"{CONSTR_ASYM} for constraint "
                                 f"term type instead of {self.constr_term} if you intend to implement "
                                 f"asymmetric uncertainties.")
    
    def get_tag_name(self):
        tag_name = f"{self.whereto}_{self.name}"
        if not self.is_common_domain():
            tag_name += f"_{self.process}"
        return tag_name
            
    def get_interp_code(self):
        constr_type = self.get_constr_type()
        return constr_type.interp_code

    def get_uncert_hi_expr(self):
        if is_float(self.errorhi):
            return f"{UNCERT_HI_PREFIX}{self.get_tag_name()}[{self.errorhi}]"
        return self.errorhi
    
    def get_uncert_lo_expr(self):
        if is_float(self.errorlo):
            return f"{UNCERT_LO_PREFIX}{self.get_tag_name()}[{self.errorlo}]"
        return self.errorlo
    
    def validate(self):
        if ((self.nominal <= 0.) and 
            (self.get_constr_type() in [ConstraintType.LogN, ConstraintType.Asym])):
            raise RuntimeError(f"failed to parse systematic {self.name}: "
                               f"constraint term {self.constr_term} received "
                               f"negative central value ({self.nominal})")


from typing import Union, Any, List, Dict, Optional

import numpy as np

def pretty_float(val:Union[str, float])->Union[int, float]:
    if float(val).is_integer():
        return int(float(val))
    return float(val)

def to_bool(val:Any):
    if not isinstance(val, str):
        return bool(val)
    else:
        if val.isdigit():
            return bool(int(val))
        else:
            if val.lower() == "true":
                return True
            elif val.lower() == "false":
                return False
            else:
                raise ValueError(f"invalid boolean expression: {val}")
                
def to_string(val:Any, precision:int=8) -> str:
    if isinstance(val, float):
        val = round(val, precision)
    return str(val)

def to_rounded_float(val:Any, precision:int=8) -> float:
    return round(float(val), precision)

def pretty_value(val:Union[int, float], precision:int=8)->Union[int, float]:
    if isinstance(val, float):
        val = round(val, precision)
        if val.is_integer():
            return int(val)
    return val

def is_integer(s:str):
    if not s:
        return False
    if len(s) == 1:
        return s.isdigit()
    if s[0] in ('-', '+'):
        return s[1:].isdigit()
    return s.isdigit()

def is_float(element: Any) -> bool:
    try:
        float(element)
        return True
    except ValueError:
        return False

def array_swap(arr1:np.ndarray, arr2:np.ndarray, indices):
    arr1[indices], arr2[indices] = arr2[indices], arr1[indices]

def df_array_swap(df, col1:str, col2:str, indices=None):
    if indices is None:
        df.loc[:, col1], df.loc[:, col2] = df[col2], df[col1]
    else:
        df.loc[indices, col1], df.loc[indices, col2] = df[indices][col2], df[indices][col1]
        
def reorder_arrays(*arrays, descending:bool=True):
    if descending:
        if not (arrays[0].dtype.type in [np.string_, np.str_]):
            indices = np.argsort(-arrays[0])
        else:
            indices = np.argsort(arrays[0])[::-1]
    else:
        indices = np.argsort(arrays[0])
    for arr in arrays:
        arr[:] = arr[indices]    

def reverse_arrays(*arrays):
    for arr in arrays:
        arr[:] = arr[::-1] 
        
def ceildiv(a, b):
    return -(-a // b)

def approx_n_digit(val:float, default=5):
    s = str(val)
    if not s.replace('.','',1).isdigit():
        return default
    elif '.' in s:
        return len(s.split('.')[1])
    else:
        return 0

def str_encode_value(val:float, n_digit=None, formatted=True):
    # account for the case where val is negative zero
    if round(float(val), 8) == 0:
        val = 0.
    if n_digit is not None:
        val_str = '{{:.{}f}}'.format(n_digit).format(val)
        #if val_str == '-{{:.{}f}}'.format(n_digit).format(0):
        #    val_str = '{{:.{}f}}'.format(n_digit).format(0)
    else:
        val_str = str(float(val))
    
    if formatted:
        val_str = val_str.replace('.', 'p').replace('-', 'n')
    return val_str

def str_decode_value(val_str):
    val = float(val_str.replace('p','.').replace('n','-'))
    return val

def is_nan_or_inf(value):
    return np.isnan(value) or np.isinf(value)

def get_bins_given_edges(low_edge:float, high_edge:float, nbins:int, decimals:int=8):
    bin_width = (high_edge - low_edge) / nbins
    low_bin_center  = low_edge + bin_width / 2
    high_bin_center = high_edge - bin_width /2
    bins = np.around(np.linspace(low_bin_center, high_bin_center, nbins), decimals)
    return bins

def array_issubset(a:np.ndarray, b:np.ndarray):
    """
    Check if array b is a subset of array a
    """
    a = np.unique(a)
    b = np.unique(b)
    c = np.intersect1d(a, b)
    return c.size == b.size

def get_proper_ranges(ranges:Union[List[float], List[List[float]]],
                      reference_value:Optional[float]=None,
                      no_overlap:bool=True):
    try:
        ranges = np.array(ranges)
    except:
        ranges = None
        
    if (ranges is None) or ranges.dtype == np.dtype('O'):
        raise ValueError("invalid range format")
    # single interval
    if ranges.ndim == 1:
        if ranges.shape == (2,):
            ranges = ranges.reshape([1, 2])
        else:
            raise ValueError("range must be array of size 2")
    if ranges.ndim != 2:
        raise ValueError("ranges must be a 2 dimensional array")
    if ranges.shape[1] != 2:
        raise ValueError("individual range must be array of size 2")
        
    ranges = ranges[ranges[:,0].argsort()]
    
    if not np.all(ranges[:, 0] <= ranges[:, 1]):
        raise ValueError("minimum range can not be greater than maximum range")
    if reference_value is not None:
        if not np.all(ranges[:, 0] <= reference_value):
            raise ValueError("minimum range is greater than the nominal value")
        if not np.all(ranges[:, 1] >= reference_value):
            raise ValueError("maximum range is smaller than the nominal value")
    if no_overlap:
        if not np.all(np.diff(ranges.flatten()) >= 0):
            raise ValueError("found overlap ranges")
    return ranges

def get_batch_slice_indices(totalsize:int, batchsize:int):
    assert (totalsize > 0) and (batchsize > 0)
    for i in range(0, totalsize, batchsize):
        yield (i, min(i + batchsize, totalsize))


from quickstats.interface.root.TObject import TObject

from quickstats.interface.root.macros import load_macros, load_macro

from quickstats.interface.root.TArrayData import TArrayData
from quickstats.interface.root.TH1 import TH1
from quickstats.interface.root.TH2 import TH2
from quickstats.interface.root.TF1 import TF1
from quickstats.interface.root.RooRealVar import RooRealVar
from quickstats.interface.root.RooAbsData import RooAbsData
from quickstats.interface.root.RooDataSet import RooDataSet
from quickstats.interface.root.RooAbsPdf import RooAbsPdf
from quickstats.interface.root.RooArgSet import RooArgSet
from quickstats.interface.root.RooMsgService import RooMsgService

load_macros()

import quickstats
quickstats.load_corelib()


from typing import Optional, Dict, List, Union
import os
import glob
import time

import numpy as np

import ROOT
import quickstats

root_type_str_maps = {
    'Char_t'    : 'char',
    'Bool_t'    : 'bool',
    'Float_t'   : 'float',
    'Double_t'  : 'double',
    'Int_t'     : 'int',
    'Long_t'    : 'long',
    'Long64_t'  : 'long long',
    'UInt_t'    : 'unsigned int',
    'ULong_t'   : 'unsigned long',
    'ULong64_t' : 'unsigned long long'
}

def rtype2ctype(rtypes:List[str]):
    ctypes = []
    for rtype in rtypes:
        #rtype = rtype.replace("ROOT::VecOps::RVec", "vector")
        ctype = root_type_str_maps.get(rtype, rtype)
        ctypes.append(ctype)
    return ctypes

def templated_rdf_snapshot(rdf:ROOT.RDataFrame, columns:List[str]=None):
    if columns is None:
        columns = [str(i) for i in rdf.GetColumnNames()]
    rtypes = [rdf.GetColumnType(c) for c in columns]
    ctypes = tuple(rtype2ctype(rtypes))
    snapshot = rdf.Snapshot.__getitem__(ctypes)
    return snapshot

def is_corrupt(f:Union[ROOT.TFile, str]):
    if isinstance(f, str):
        f = ROOT.TFile(f)
    if f.IsZombie():
        return True
    if f.TestBit(ROOT.TFile.kRecovered):
        return True
    if f.GetNkeys() == 0:
        return True
    return False

def release_tfiles():
    for f in ROOT.gROOT.GetListOfFiles():
        f.Close()

def compile_macro(name:str, opt:str="kfOc"):
    macros_dir = os.path.join(quickstats.macro_path, name)
    macros_path = os.path.join(macros_dir, name + '.cxx')
    print('INFO: Compiling macro "{}"...'.format(name))
    if not os.path.exists(macros_path):
        print('ERROR: Cannot locate macro files from {}. ROOT macros will not be compiled.'.format(macros_dir))
        return -1
    status = ROOT.gSystem.CompileMacro(macros_path, opt)
    #status = ROOT.gROOT.LoadMacro('{}++'.format(macros_path))
    return status

def load_macro(name:str):
    import cppyy
    if name in dir(cppyy.gbl):
        return None
    auto_load = ROOT.gInterpreter.AutoLoad(name)
    # already loaded by quickstats / third-party
    if auto_load == 1:
        return None
    class_loaded = hasattr(ROOT, name)
    if class_loaded:
        return None
    macros_dir = os.path.join(quickstats.macro_path, name)
    macros_path = os.path.join(macros_dir, name + '_cxx')
    if not os.path.exists(macros_path + '.so'):
        status_compile = compile_macro(name)
    status_load = ROOT.gSystem.Load(macros_path)
    if (status_load != 0) and (status_load != 1):
        raise RuntimeError(f"Shared library for the macro {name} is incompatible with the current pyROOT version. "
                           "Please recompile by using the command line \"quickstats compile\". "
                           "(use quickstats.compile_macros() instead inside a jupyter kernel)")
    return status_load

def get_macro_dir(name:str):
    return os.path.join(quickstats.macro_path, name)

def get_macro_TClass(name:str):
    if not getattr(ROOT, name):
        raise RuntimeError(f"class {name} is not loaded into ROOT yet")    
    cls = ROOT.TClass.GetClass(name)
    macro_dir = get_macro_dir(name)
    macro_decl_file = os.path.join(f"{macro_dir}/{name}.h")
    macro_impl_file = os.path.join(f"{macro_dir}/{name}.cxx")
    if not os.path.exists(macro_decl_file):
        raise FileNotFoundError(f"decl file {macro_decl_file} for the macro \"{name}\" not found")
    if not os.path.exists(macro_impl_file):
        raise FileNotFoundError(f"impl file {macro_impl_file} for the macro \"{name}\" not found")        
    cls.SetDeclFile(macro_decl_file, 0)
    cls.SetImplFileName(macro_impl_file)
    return cls

def print_resource_used(start):
    proc_info = ROOT.ProcInfo_t()
    flag = ROOT.gSystem.GetProcInfo(proc_info)
    if flag < 0:
        return None
    cpu_time = proc_info.fCpuUser + proc_info.fCpuSys
    wall_time = time.time()-start
    print('Resources Used: cpu_time={:.3f}s, mem={}kb, vmem={}kb, wall_time={:.3f}s'.format(
          cpu_time, proc_info.fMemResident, proc_info.fMemVirtual, wall_time))
    return None

def process_uproot_array(source):
    result = np.array(source)
    if (result.dtype == np.dtype('O')) and (isinstance(result[0], bytes)):
        result = result.astype(str)
    return result

def get_data_size(data):
    if isinstance(data, (tuple, list, np.ndarray)):
        return len(data)
    else:
        return 1
    
def root_to_dict(filename:str, simplify:bool=True):
    f = ROOT.TFile(filename)
    if is_corrupt(f):
        raise RuntimeError(f"corrupted ROOT file: {filename}")
    treenames = [t.GetName() for t in f.GetListOfKeys()]
    result = {}
    for treename in treenames:
        tree = f.Get(treename)
        rdf = ROOT.RDataFrame(tree)
        data = rdf.AsNumpy()
        for branch in data:
            # convert std::string to python string
            if rdf.GetColumnType(branch) == "string":
                data[branch] = np.char.mod('%s', data[branch])
            # unwrap array if size is 1
            if simplify and (data[branch].shape[0] == 1):
                data[branch] = data[branch][0]
        result[treename] = data
    return result    

def uproot_to_dict(file):
    result = {}
    for tree in file.values():
        #tree_name = tree.name.decode('utf-8')
        tree_name = tree.name
        result[tree_name] = {}
        for branch in tree.values():
            #branch_name = branch.name.decode('utf-8')
            branch_name = branch.name
            arr = process_uproot_array(branch.array())
            value = arr[0] if len(arr) == 1 else arr
            result[tree_name][branch_name] = value    
    return result
    
def get_leaf_type_str(data):
    from six import string_types
    
    type_str = None
    data_type = None

    if isinstance(data, (tuple, list, np.ndarray)):
        data_type = type(data[0])
        if data_type in string_types:
            return None
    elif isinstance(data, string_types):
        return None
    else:
        data_type = type(data)
        
    if data_type in [float, np.float64]:
        type_str = 'D'
    elif data_type in [int, np.int64]:
        type_str = 'L'
    elif data_type in [np.int32]:
        type_str = 'I'
    elif data_type in [np.uint32]:
        type_str = 'i'        
    elif data_type in [bool, np.bool_]:
        type_str = 'O'        
    elif data_type in [np.float32]:
        type_str = 'F'
    elif data_type in [np.int8]:
        type_str = 'B'
    elif data_type in [np.uint8]:
        type_str = 'b'
    else:
        raise ValueError('cannot infer leaf type for {}'.format(data_type))
    return type_str

def fill_branch(tree, values_dict):
    buffer = {}
    leaf_types = {}
    data_sizes = list(set([get_data_size(v) for k,v in values_dict.items()]))
    if len(data_sizes) > 1:
        raise ValueError('cannot fill tree with values of inconsistent entry sizes')
    data_size = data_sizes[0]
    if data_size == 1:
        for k, v in values_dict.items():
            leaf_type_str = get_leaf_type_str(v)
            leaf_types[k] = leaf_type_str
            if leaf_type_str is not None:
                buffer[k] = np.array([v], dtype=type(v))
                tree.Branch(k, buffer[k], '{}/{}'.format(k, leaf_type_str))
            else:
                buffer[k] = ROOT.std.string(v)
                tree.Branch(k, buffer[k])
        tree.Fill()
    else:
        for k, v in values_dict.items():
            leaf_type_str = get_leaf_type_str(v)
            leaf_types[k] = leaf_type_str
            if leaf_type_str is not None:
                buffer[k] = np.array([v[0]], dtype=type(v[0]))
                tree.Branch(k, buffer[k], '{}/{}'.format(k, leaf_type_str))
            else:
                buffer[k] = ROOT.std.string(v[0])
                tree.Branch(k, buffer[k])
        for i in range(data_size):
            for k, v in values_dict.items():
                if leaf_types[k] is not None:
                    buffer[k][0] = v[i]
                else:
                    buffer[k].replace(0, ROOT.std.string.npos, v[i])
            tree.Fill()
            
def close_all_root_files():
    opened_files = ROOT.gROOT.GetListOfFiles()
    for f in opened_files:
        f.Close()
        
def create_declaration(expression:str, name:Optional[str]=None):
    if name is None:
        hash_str = str(hash(expression)).replace("-", "n")
        name_guard = f"__quickstats_declare_{hash_str}__"
    else:
        name_guard = f"__quickstats_declare_{name}__"
    guarded_declaration = f"#ifndef {name_guard}\n"
    guarded_declaration += f"#define {name_guard}\n"
    guarded_declaration += f"\n{expression}\n\n#endif\n"
    return guarded_declaration

def declare_expression(expression:str, name:Optional[str]=None):
    declaration = create_declaration(expression, name)
    status = ROOT.gInterpreter.Declare(declaration)
    return status

def get_tree_names(f:ROOT.TFile):
    tree_names = []
    for i in f.GetListOfKeys():
        obj = f.Get(i.GetName())
        if isinstance(obj, ROOT.TTree):
            tree_names.append(obj.GetName())
    return tree_names

def merge_root_files(source_files:Union[str, List[str]], target_file:str, treename:str=None,
                     multithread:bool=True, keep_order:bool=False, use_template:bool=True,
                     ignore_missing_columns:bool=True, verbosity:str="INFO"):
    from quickstats.utils.io import VerbosePrint
    stdout = VerbosePrint(verbosity)
    
    if isinstance(source_files, str):
        source_files = [source_files]
    rfiles = []
    for source_file in source_files:
        files = glob.glob(source_file)
        if not files:
            raise FileNotFoundError(f"file \"{source_file}\" does not exist")
        rfiles += files
    # now all rfiles are guaranteed to exist
    for i, rfile in enumerate(rfiles):
        stdout.info(f"Source {i+1}: {rfile}")
        if is_corrupt(rfile):
            raise RuntimeError(f"the root file \"{rfile}\" is corrupted")
    if not rfiles:
        raise RuntimeError("no source files found")
    if treename is None:
        f = ROOT.TFile(rfiles[0])
        treenames = get_tree_names(f)
        if len(treenames) > 1:
            raise RuntimeError("unable to deduce tree name from source files: "
                               "multiple trees are present")
        treename = treenames[0]
        stdout.info("INFO: No tree name specified for merging. Will use "
                    f"the auto-detected tree name \"{treename}\".")
        
    # configure multi-threading setting
    multithread_orig_state = ROOT.IsImplicitMTEnabled()
    if multithread and (not multithread_orig_state) and (not keep_order):
        ROOT.EnableImplicitMT()
    if keep_order and multithread_orig_state:
        ROOT.DisableImplicitMT()
    
    if ignore_missing_columns:
        column_map = {}
        all_columns = set()
        for rfile in rfiles:
            rdf = ROOT.RDataFrame(treename, rfile)
            columns = set([str(i) for i in rdf.GetColumnNames()])
            all_columns.update(columns)
            column_map[rfile] = columns
        for fname, columns in column_map.items():
            if columns != all_columns:
                missing_columns = list(all_columns - columns)
                stdout.warning(f"WARNING: The root file \"{rfile}\" is missing the following column(s): "
                               f"{','.join(missing_columns)}. The missing column(s) will not be saved.")
        columns_to_keep = list(set.intersection(*column_map.values()))
        rdf = ROOT.RDataFrame(treename, rfiles)
        stdout.info(f"Target path: {target_file}")
        if use_template:
            snapshot = templated_rdf_snapshot(rdf, columns_to_keep)
        else:
            snapshot = rdf.Snapshot
        snapshot(treename, target_file, columns_to_keep)
    else:
        rdf = ROOT.RDataFrame(treename, rfiles)
        stdout.info(f"Target path: {target_file}")
        if use_template:
            snapshot = templated_rdf_snapshot(rdf, columns_to_keep)
        else:
            snapshot = rdf.Snapshot
        snapshot(treename, target_file)
   
    # restore multi-threading setting
    if multithread_orig_state != ROOT.IsImplicitMTEnabled():
        if multithread_orig_state:
            ROOT.DisableImplicitMT()
        else:
            ROOT.EnableImplicitMT() 


from .basics import *
from .root_object import ROOTObject
from .extended_minimizer import ExtendedMinimizer
from .extended_model import ExtendedModel
from .analysis_object import AnalysisObject
from .asimov_generator import AsimovGenerator
from .asimov_generator import AsimovType
from .nuisance_parameter_pull import NuisanceParameterPull
from .nuisance_parameter_harmonizer import NuisanceParameterHarmonizer
from .asymptotic_cls import AsymptoticCLs
from .likelihood import Likelihood
from .pvalue_toys import PValueToys
from .toy_limit_calculator import ToyLimitCalculator
from .analysis_base import AnalysisBase
from .roo_inspector import RooInspector
from .extended_rfile import ExtendedRFile
#from .signal_modelling import SignalModelling

import ROOT
ROOT.gROOT.SetBatch(True)


from typing import Optional, Union, List, Dict, Tuple
import os
import math
import json
from itertools import combinations, combinations_with_replacement

import numpy as np
import pandas as pd

from quickstats import semistaticmethod, DescriptiveEnum, Timer
from quickstats.maths.numerics import str_encode_value
from quickstats.maths.statistics import (get_poisson_interval, get_counting_significance,
                                         get_combined_counting_significance)
from quickstats.utils.common_utils import is_valid_file, combine_dict
from quickstats.parsers import ParamParser
from .data_loading import DataLoader
from .multi_class_boundary_tree import MultiClassBoundaryTree
from .data_preprocessing import get_class_label_encodings

class CategorizationOutputMode(DescriptiveEnum):
    MINIMAL  = (0, "Save only outputs needed for statistical evaluation")
    MINITREE = (1, "Save minitrees for all test samples")
    ARRAY    = (2, "Save arrays(csv/h5) for all test samples")
    STANDARD = (3, "Save arrays(csv/h5) and minitrees for all test samples")
    FULL     = (4, "Save arrays(csv/h5), histograms and minitrees for all test samples")
    
class CategorizationOutputFormat(DescriptiveEnum):
    ARRAY     = (0, "array file (csv/h5)", "categorized_array", "categorized_array_sample", "array")
    MINITREE  = (1, "ROOT file containing minitrees", "categorized_minitree", "categorized_minitree_sample", "minitree")
    HISTOGRAM = (2, "ROOT file containing histograms", "categorized_histogram", "categorized_histogram_sample", "histogram")
    
    def __new__(cls, value:int, description:str="", dirname:str="", filename:str="", short_name:str=""):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.description = description
        obj.dirname = dirname
        obj.filename = filename
        obj.short_name = short_name
        return obj

class EventCategorization(DataLoader):
    
    REQUIRED_CONFIG_COMPONENTS = ["paths:arrays", "paths:model", "paths:output",
                                  "samples:*", "kinematic_regions:*",
                                  "variables:*", "training:*", "channels:*",
                                  "observables:*", "categorization:*",
                                  "benchmark:*", "names:weight", "names:event_number"]
    
    def __init__(self, study_name:str, target_channels:List[str],
                 analysis_config:Optional[Union[Dict, str]]=None,
                 array_dir:Optional[str]=None, outdir:Optional[str]=None,
                 data_preprocessor=None, do_blind:bool=True,
                 verbosity:Optional[Union[int, str]]="INFO",
                 **kwargs):
        """
        Module for performing event categorization based on score outputs from the machine learning
        model.
        
        Parameters
        ----------
        study_name: str
            Name of the study (for grouping outputs)
        target_channels: list of str
            List of channels to be included in the study
        analysis_config_path: str
            Path to the central configuration file of the analysis
        outdir: (optional) str
            Save outputs to a custom directory instead of the default directory
            from the configuration file (i.e. config['paths']['outputs'])
        array_dir: (optional) str
            Load csv array inputs from a custom directory instead of the default
            directory from the configuration file (i.e. config['paths']['arrays'])
        data_preprocessor: callable function
            Function for preprocessing input data after which the output is
            fed to the machine learning model for prediction
        do_blind: bool
            Perform a blind analysis. This will remove data events from the signal region when
            saving outputs for statistical tests.
        verbosity: (optional) str
            Message verbosity level. Available chocie: ['DEBUG', 'INFO', 'WARNING', 'ERROR']
        """
        super().__init__(analysis_config=analysis_config,
                         array_dir=array_dir,
                         outdir=outdir,
                         verbosity=verbosity,
                         **kwargs)
        
        # append the study name to the output directory
        self.set_study_name(study_name)
        self.do_blind  = do_blind
        self.data_preprocessor = data_preprocessor
        self.models = None
        self.initialize_channels(target_channels)
        self.reset_channel_information()
        self.boundary_trees = {}
        self.summary    = {}
        self.category_yields     = {}
        self.category_yields_err = {}
        self.benchmark_significance = {}
    
    def reset_channel_information(self) -> None:
        self.active_channel = None
        self.channel_df     = None
        self.channel_config = None
        
    def set_active_channel(self, channel:str) -> None:
        if channel not in self.target_channels:
            raise ValueError(f"{channel} is not a target channel of this study")
        self.reset_channel_information()
        self.active_channel = channel
        self.channel_config = self.channel_configs[channel]
        
    def set_models(self, models:Dict)-> None:
        """
        Define machine learning models used for categorization.
        
        Parameters
        ----------
        models: dictionary
            A dictionary with key as the channel name and value as the machine learning model used
            to evaluate event-level scores.
        """
        channel_models = {}
        for channel in self.target_channels:
            model = models.get(channel, None)
            if model is None:
                raise ValueError(f"missing model for the channel: {channel}")
            channel_models[channel] = model
        self.models = channel_models
        
    def validate(self, check_model:bool=False, check_df:bool=False,
                 check_boundaries:bool=False, check_categories:bool=False,
                 check_channels:bool=False) -> None:
        """
        Check if inputs are available for various steps in the categorization procedure.

        Parameters
        ----------
        check_model: bool, default = False
            Check whether machine learning models are defined for the active channel.
        check_df: bool, default = False
            Check whether the channel-level sample dataframes are loaded.
        check_boundaries: bool, default = False
            Check whether the channel-level boundaries are defined.
        check_categories: bool, default = False
            Check whether the channel-level categories are defined.
        check_channels: bool, default = False
            Check whether boundaries for all channels are defined
        """
        if self.active_channel is None:
            raise RuntimeError("active channel not set")
        if check_model and ((self.models is None) or (self.models.get(self.active_channel, None) is None)):
            raise RuntimeError(f'model for the channel "{self.active_channel}" not set')
        if check_df and (not self.channel_df):
            raise RuntimeError(f"channel dataframes not initialized")
        if check_boundaries and (self.active_channel not in self.boundary_trees):
            raise RuntimeError(f"channel score boundaries not defined")
        if check_categories:
            class_labels = self.get_class_labels()
            columns = [self.get_category_column_name(class_label) for class_label in class_labels]
            for sample in self.channel_df:
                if not set(columns).issubset(self.channel_df[sample].columns):
                    raise RuntimeError("category columns not defined")
        if check_channels:
            category_names = self.get_all_category_names(valid_only=True)
            involved_channels = set([self.get_category_channel(category_name) for category_name in category_names])
            missing_channels = set(self.target_channels) - involved_channels
            if len(missing_channels) > 0:
                raise RuntimeError(f'missing boundary information for the channel(s): {", ".join(missing_channels)}')
    
    def resolve_train_samples(self, channel:str) -> List[str]:
        channel_config  = self.config['channels'][channel]
        train_samples   = self.resolve_samples(channel_config['train_samples'])
        # add samples that are indirectly used in the training
        indirect_train_samples = self.resolve_samples(channel_config.get('indirect_train_samples', []))
        train_samples.extend(indirect_train_samples)
        return train_samples
    
    def initialize_channels(self, target_channels:List[str], **kwargs) -> None:
        """
        Initializes channel-level information used in the categorization procedure.

        Parameters
        ----------
        target_channels: list of string
            Channels to perform the initialization.
        """
        
        bc_ds = self.config["categorization"]["boundary_scan"].get("datasets", None)
        eval_ds = self.config["categorization"]["evaluation"].get("datasets", None)
        if bc_ds is not None:
            self.stdout.info("INFO: Boundary scan will be performed on the following analysis "
                            f"dataset(s): {', '.join(bc_ds)}")
        else:
            self.stdout.info("INFO: Boundary scan will be performed on the full analysis dataset.")
        if eval_ds is not None:
            self.stdout.info("INFO: Statistical evaluation will be performed on the following analysis "
                            f"dataset(s): {', '.join(eval_ds)}")
        else:
            self.stdout.info("INFO: Statistical evaluation will be performed on the full analysis dataset.")
            
        self.target_channels = self.resolve_channels(target_channels)
        channel_configs  = {}
        channel_metadata = {}
        for channel in self.target_channels:
            self.stdout.info(f'INFO: Initializing configuration for the channel "{channel}"')
            channel_config  = self.config['channels'][channel]
            train_variables = self.resolve_variables(channel_config['train_variables'])
            train_samples   = self.resolve_train_samples(channel)
            test_samples    = self.resolve_samples(channel_config['test_samples'])
            num_class = len(channel_config["class_labels"])
            # check that class labels are valid
            if num_class == 2:
                class_labels = set(channel_config["class_labels"].keys())
                if class_labels != set(["0", "1"]):
                    raise RuntimeError('binary classification must have class labels "0" (background) '
                                       'and "1" (signal)')
            elif num_class < 2:
                raise RuntimeError("number of training classes can not be less than 2")
            class_labels = channel_config['class_labels']
            class_label_encodings = get_class_label_encodings(list(class_labels.keys()))
            # set up score labels, category labels for the given classes
            class_metadata = {}
            category_metadata = {}
            if num_class > 2:
                buffer = []
                for class_label in channel_config["counting_significance"]:
                    class_metadata[class_label] = {
                        "score_column_name"    : f'{class_label}_{self.names["score"]}',
                        "category_column_name" : f'{class_label}_{self.names["category"]}',
                        "encoded_index"        : class_label_encodings[class_label],
                        "counting_significance": channel_config["counting_significance"][class_label]
                    }
                    buffer_new = []
                    n_boundaries = class_metadata[class_label]["counting_significance"]["n_boundaries"]
                    for i in range(n_boundaries + 1):
                        if len(buffer) == 0:
                            index = tuple([i])
                            label = f"{class_label}_{i}"
                            category_metadata[index] = {
                                "label": label,
                                "valid": False
                            }
                            buffer_new.append((label, index))
                        else:
                            for j in buffer:
                                index = tuple(list(j[1]) + [i])
                                label = f"{j[0]}_{class_label}_{i}"
                                category_metadata[index] = {
                                    "label": label,
                                    "valid": False
                                }
                                buffer_new.append((label, index))
                    buffer.extend(buffer_new)
            # binary classification, single score                
            else:
                class_metadata[None] = {
                    "score_column_name"    : self.names["score"],
                    "category_column_name" : self.names["category"],
                    "encoded_index"        : None,
                    "counting_significance": channel_config["counting_significance"]
                }
                n_boundaries = class_metadata[None]["counting_significance"]["n_boundaries"]
                for i in range(n_boundaries + 1):
                    category_metadata[i] = {
                        "label": str(i),
                        "valid": False
                    }
            
            channel_configs[channel] = {
                "train_variables"   : train_variables,
                "train_samples"     : train_samples,
                "test_samples"      : test_samples,
                "num_class"         : num_class,
                "selection"         : channel_config.get("selection", None),
                "kinematic_region"  : channel_config.get("kinematic_region", None),
                "exclude_categories": channel_config["exclude_categories"]
            }

            channel_metadata[channel] = {
                "classes"   : class_metadata,
                "categories": category_metadata
            }
        self.channel_configs  = channel_configs
        self.channel_metadata = channel_metadata
    
    def _parse_channel(self, channel:Optional[str]=None):
        if channel is None:
            self.validate()
            channel = self.active_channel
        elif channel not in self.target_channels:
            raise RuntimeError(f'channel "{channel}" is not a target channel of this study')
        return channel
        
    def get_boundary_tree(self, channel:Optional[str]=None) -> MultiClassBoundaryTree:
        """
        Get boundary data in a given channel.
        
        Parameters
        ----------
        channel: (optional) str
            The channel from which the boundary data are retrieved.
            By default, the current channel is used.
        """        
        channel = self._parse_channel(channel)
        if channel not in self.boundary_trees:
            raise RuntimeError(f'boundaries not initialized for the channel "{channel}"')
        return self.boundary_trees[channel]
    
    def get_class_labels(self, channel:Optional[str]=None) -> List[str]:
        """
        Get class labels in a given channel.
        
        Parameters
        ----------
        channel: (optional) str
            The channel from which the class labels are retrieved.
            By default, the current channel is used.
        """
        channel = self._parse_channel(channel)
        class_labels = list(self.channel_metadata[channel]["classes"].keys())
        return class_labels
    
    def get_class_metadata(self, class_label:Optional[str]=None,
                           channel:Optional[str]=None) -> Dict:
        """Get metadata of a class defined in machine learning training in a given channel.
        
        Arguments:
            class_label: (optional) str
                Name of the class defined in machine learning training.
            channel: (optional) str
                The channel from which the class_label is retrieved.
                By default, the current channel is used.
        """
        channel = self._parse_channel(channel)
        class_metadata = self.channel_metadata[channel]["classes"]
        if class_label not in class_metadata:
            raise ValueError(f'undefined class label: "{class_label}"')
        return class_metadata[class_label]
        
    def get_score_column_name(self, class_label:Optional[str]=None,
                              channel:Optional[str]=None) -> str:
        """Get score column name associated with the class label in a given channel.
        
        Arguments:
            class_label: (optional) str
                The class label for which the corresponding score column name should be obtained.
                In case of binary classification, the class label can be omitted.
            channel: (optional) str
                The channel from which the score column name is retrieved.
                By default, the current channel is used.
        """
        class_metadata = self.get_class_metadata(class_label=class_label,
                                                 channel=channel)
        return class_metadata["score_column_name"]
    
    def get_category_column_name(self, class_label:Optional[str]=None,
                                 channel:Optional[str]=None) -> str:
        """
        Get category column name associated with the class label in a given channel.
        
        Parameters
        ----------
        class_label: str
            The class label for which the corresponding category column name should be obtained.
            In case of binary classification, the class label can be omitted.
        channel: (optional) str
            The channel from which the category column name is retrieved.
            By default, the current channel is used.
        """
        class_metadata = self.get_class_metadata(class_label=class_label,
                                                 channel=channel)
        return class_metadata["category_column_name"]
    
    def get_category_label(self, index:Union[int, Tuple[int], List[int]],
                           channel:Optional[str]=None) -> str:
        """
        Get category label associated with the category index in a given channel.
        
        Parameters
        ----------
        index: int, tuple of int, list of int
            The category index that the category label is associated with.
        channel: (optional) str
            The channel from which the category label is retrieved.
            By default, the current channel is used.
        """
        
        channel = self._parse_channel(channel)
        category_metadata = self.channel_metadata[channel]["categories"]
        if isinstance(index, list):
            index = tuple(index)
        if index not in category_metadata:
            raise RuntimeError(f'no category with index "{index}" found in the channel "{channel}"')
        return category_metadata[index]["label"]
    
    def get_category_name(self, index_or_label:Union[str, int, Tuple[int], List[int]],
                          channel:Optional[str]=None) -> str:
        """
        Get category name associated with the category index/label in a given channel.
        
        Parameters
        ----------
        index: str, int, tuple of int, list of int
            The category index/label that the category name is associated with.
        channel: (optional) str
            The channel from which the category name is retrieved.
            By default, the current channel is used.
        """        
        channel = self._parse_channel(channel)
        if isinstance(index_or_label, str):
            category_label = index_or_label
        else:
            category_label = self.get_category_label(index_or_label, channel=channel)
        category_name  = f"{channel}_{category_label}"
        return category_name
    
    def get_category_index(self, label:str, channel:Optional[str]=None):
        """
        Get category index associated with the category label in a given channel.
        
        Parameters
        ----------
        label: str
            The category label that the category index is associated with.
        channel: (optional) str
            The channel from which the category index is retrieved.
            By default, the current channel is used.
        """           
        channel = self._parse_channel(channel)
        category_metadata = self.channel_metadata[channel]["categories"]
        for index, data in category_metadata.items():
            if (data["label"] == label) or (f'{channel}_{data["label"]}' == label):
                return index
        raise RuntimeError(f'no category with label "{label}" found in the channel "{channel}"')
        
    def get_category_channel(self, category_name:str):
        """
        Get channel associated with the given category name.
        
        Parameters
        ----------
        category_name: str
            The category name that the channel is associated with.
        """
        for channel in self.target_channels:
            category_metadata = self.channel_metadata[channel]["categories"]
            for index in category_metadata:
                if self.get_category_name(index, channel=channel) == category_name:
                    return channel
        raise RuntimeError(f'no category named "{category_name}" found')
        
    def get_category_labels(self, channel:Optional[str]=None, valid_only:bool=True) -> List[str]:
        """
        Get all category labels in a given channel.
        
        Parameters
        ----------
        channel: (optional) str
            The channel from which the category labels are retrieved.
            By default, the current channel is used.
        valid: bool, default = True
            Whether to filter invalid categories.
        """
        channel = self._parse_channel(channel)
        category_metadata = self.channel_metadata[channel]["categories"]
        if valid_only:
            return [data["label"] for data in category_metadata.values() if data["valid"]]
        return [data["label"] for data in category_metadata.values()]
    
    def get_category_indices(self, channel:Optional[str]=None,
                             valid_only:bool=True) -> List[Union[int, Tuple[int]]]:
        """
        Get all category indices in a given channel.
        
        Parameters
        ----------
        channel: (optional) str
            The channel from which the category indices are retrieved.
            By default, the current channel is used.
        valid: bool, default = True
            Whether to filter invalid categories.
        """
        channel = self._parse_channel(channel)
        category_metadata = self.channel_metadata[channel]["categories"]
        if valid_only:
            return [index for index, data in category_metadata.items() if data["valid"]]
        return list(category_metadata.keys())
    
    def get_category_names(self, channel:Optional[str]=None,
                           valid_only:bool=True) -> List[str]:
        """
        Get all category names in a given channel.
        
        Parameters
        ----------
        channel: (optional) str
            The channel from which the category names are retrieved.
            By default, the current channel is used.
        valid_only: bool, default = True
            Whether to filter invalid categories.
        """
        channel = self._parse_channel(channel)
        categories = []
        category_labels = self.get_category_labels(channel=channel,
                                                   valid_only=valid_only)
        category_names = [f"{channel}_{category_label}" for category_label in category_labels]
        return category_names
    
    def get_all_category_names(self, valid_only:bool=True) -> List[str]:
        """
        Get category names in all target channels.
        
        Parameters
        ----------
        valid_only: bool, default = True
            Whether to filter invalid categories.
        """
        category_names = []
        for channel in self.target_channels:
            category_names += self.get_category_names(channel=channel,
                                                      valid_only=valid_only)
        return category_names
    
    def get_columns_to_load(self, channel:Optional[str]=None):
        """
        Get the list of columns to load from data arrays.
        
        Parameters
        ----------
        channel: (optional) str
            The channel from which the columns are retrieved.
            By default, the current channel is used.
        """
        channel = self._parse_channel(channel)
        columns = set()
        for column in ["event_number", "weight"]:
            if (column in self.names) and (self.names[column] is not None):
                columns |= set([self.names[column]])
        columns |= set(self.config["categorization"].get("save_variables", []))
        columns |= set(self.channel_configs[channel]["train_variables"])
        # columns needed for selecting analysis datasets
        for spec in self.config["training"]["datasets"]["specification"].values():
            dataset_selection_variables = spec.get("variables", [])
            columns |= set(dataset_selection_variables)
        channel_selection_variables = self.channel_configs[channel].get("selection_variables", None)
        if channel_selection_variables is not None:
            columns |= set(channel_selection_variables)
        return list(columns)
    
    def get_columns_to_save(self, channel:Optional[str]=None):
        """
        Get the list of columns to be saved in array/minitree outputs.
        
        Parameters
        ----------
        channel: (optional) str
            The channel from which the columns are retrieved.
            By default, the current channel is used.
        """
        channel = self._parse_channel(channel)
        columns = set()
        for column in ["event_number", "weight", "total_weight"]:
            if (column in self.names) and (self.names[column] is not None):
                columns |= set([self.names[column]])
        columns |= set(self.config["categorization"].get("save_variables", []))
        # score variables and category variables
        class_labels     = self.get_class_labels(channel)
        score_columns    = [self.get_score_column_name(class_label) for class_label in class_labels]
        category_columns = [self.get_category_column_name(class_label) for class_label in class_labels]
        columns |= set(score_columns)
        columns |= set(category_columns)
        if ("observable" in self.config) and ("name" in self.config["observable"]):
            columns |= set([self.config["observable"]["name"]])
        return list(columns)    
  
    def get_analysis_dataset_weight_expr(self, dataset_names:Union[List[str], str]):
        if isinstance(dataset_names, str):
            dataset_names = [dataset_names]
        effective_weight = 0
        weight_components = []
        dataset_specs = self.config["training"]["datasets"]["specification"]
        for dataset_name in dataset_names:
            if dataset_name not in dataset_specs:
                raise RuntimeError(f'analysis dataset "{dataset_name}" not defined')
            ds_selection = dataset_specs[dataset_name]["selection"]
            ds_effective_weight = dataset_specs[dataset_name]["effective_weight"]
            weight_components.append(ds_selection)
            effective_weight += ds_effective_weight
        weight_expr = " or ".join([f"({c})" for c in weight_components])
        weight_expr = f"(({weight_expr}) * {1 / effective_weight})"
        return weight_expr
    
    def get_blind_selection(self):
        # analysis specific, should be overriden
        return "1"
    
    def get_sample_weight_expr(self, sample:str, for_boundary_scan:bool=True,
                               ignore_dataset_weight:bool=False) -> str:
        """
        Get expression for filtering and weighting events.
            
        Parameters
        ----------
        sample: str
            Name of sample.
        for_boundary_scan: boolean, default = True
            Whether the expression is used for boundary scan.
        ignore_dataset_weight: boolean, default = False
            Whether to ignore weight correction from using specific analysis datasets.
            Only relevant when `for_boundary_scan` is set to True.
        """
        weight_components = []
        weight_components.append(self.names["weight"])
        train_samples = self.channel_config["train_samples"]
        if for_boundary_scan:
            # boundary scan specific selection
            bs_config = self.config["categorization"]["boundary_scan"]
            selection = bs_config.get("selection", {})
            sample_selection = selection.get(sample, None)
            if sample_selection is not None:
                weight_components.append(sample_selection)
            bs_datasets = bs_config.get("datasets", [])
            if bs_datasets and (not ignore_dataset_weight):
                if (not bs_config["adaptive_dataset"]) or (sample in train_samples):
                    weight_expr = self.get_analysis_dataset_weight_expr(bs_datasets)
                    weight_components.append(weight_expr)
        else:
            # statistical evaluation specific selection
            eval_config = self.config["categorization"]["evaluation"]
            selection = eval_config.get("selection", {})
            sample_selection = selection.get(sample, None)
            if sample_selection is not None:
                weight_components.append(sample_selection)
            eval_datasets = eval_config.get("datasets", [])
            if eval_datasets:
                if (not eval_config["adaptive_dataset"]) or (sample in train_samples):
                    weight_expr = self.get_analysis_dataset_weight_expr(eval_datasets)
                    weight_components.append(weight_expr)
            # remove data events in the signal region in case of blind analysis
            if self.is_data_sample(sample) and self.do_blind:
                blind_selection = self.get_blind_selection()
                weight_components.append(blind_selection)
        weight_expr = " * ".join(weight_components)
        # convert c expression to python expression
        weight_expr = weight_expr.replace("&&", "&")
        weight_expr = weight_expr.replace("!", "not")
        return weight_expr
    
    def apply_observable(self, df:pd.DataFrame):
        name      = self.config["observable"]["name"]
        if (name not in df.columns):
            eval_expr = self.config["observable"].get("eval_expr", None)
            if eval_expr is None:
                raise RuntimeError(f'Failed evaluate observable "{name}": no evaluation expression available. '
                                    'Please specify the expression in config["observable"]["eval_expr"]')
            df[name] = df.eval(eval_expr)
            
    def initialize_df(self, df:pd.DataFrame, sample:str,
                      apply_score:bool=True,
                      event_indices:Optional[np.ndarray]=None):
        
        # evaluate observable if not already exists
        self.apply_observable(df)
        
        use_custom_events = event_indices is not None
        weight_name = self.names["weight"]
        # weight for categorization
        cat_weight_expr = self.get_sample_weight_expr(sample, for_boundary_scan=True,
                                                      ignore_dataset_weight=use_custom_events)
        cat_weight_name = self.names["cat_weight"]
        if event_indices is None:
            df[cat_weight_name] = df.eval(cat_weight_expr)
        # validation data from cross-validation
        else:
            index_variable = self.get_event_index_variable()
            df.set_index(index_variable, inplace=True)
            df[cat_weight_name] = 0.
            # scale from validation dataset to inclsuive dataset
            inclusive_weight = df[weight_name].sum()
            selected_weight = (df.loc[event_indices])[weight_name].sum()
            weight_scale = inclusive_weight / selected_weight
            cat_weight = df.loc[event_indices].eval(cat_weight_expr) * weight_scale
            df.loc[event_indices, cat_weight_name] = cat_weight
            df.reset_index(inplace=True)

        # weight for stat analysis
        total_weight_expr = self.get_sample_weight_expr(sample, for_boundary_scan=False)
        total_weight_name = self.names["total_weight"]
        df[total_weight_name] = df.eval(total_weight_expr)

        # evaluate MVA score
        if apply_score:
            self.apply_score(df)

    def get_samples_to_load(self, channel:Optional[str]=None):
        channel = self._parse_channel(channel)
        channel_config = self.channel_configs[channel]
        return channel_config['test_samples']
    
    def set_boundary_scan_dataset_options(self, datasets:Optional[str]=None,
                                          adaptive:bool=True):
        self.config["categorization"]["boundary_scan"]["datasets"] = boundary_scan
        self.config["categorization"]["boundary_scan"]["adaptive_dataset"] = adaptive
        
    def set_evaluation_dataset_options(self, datasets:Optional[str]=None,
                                       adaptive:bool=True):
        self.config["categorization"]["evaluation"]["datasets"] = boundary_scan
        self.config["categorization"]["evaluation"]["adaptive_dataset"] = adaptive
    
    def load_channel_df(self, samples:Optional[List]=None, apply_score:bool=True,
                        event_indices:Optional[Dict[str, np.ndarray]]=None) -> None:
        if apply_score:
            self.validate(check_model=True)
        else:
            self.validate(check_model=False)
        channel = self.active_channel
        channel_selection = self.channel_config['selection']
        kinematic_region  = self.channel_config['kinematic_region']
        self.stdout.info(f"INFO: Loading inputs for the channel: {channel}")
        if samples is None:
            samples = self.get_samples_to_load()
            
        channel_df = {}
        for sample in samples:
            sample_df = self.get_sample_df(sample, columns=None,
                                           selection=channel_selection,
                                           kinematic_region=kinematic_region)
            if event_indices is not None:
                sample_event_indices = event_indices.get(sample, None)
            else:
                sample_event_indices = None
            self.initialize_df(sample_df, sample=sample,
                               apply_score=apply_score,
                               event_indices=sample_event_indices)
            channel_df[sample] = sample_df
        self.channel_df = channel_df
    
    def apply_score(self, df:pd.DataFrame) -> None:
        """
        Evaluate score value from the machine learning model using events from a given dataframe and
        append the score to a new column in the dataframe.

        Parameters
        ----------
        df: pandas.Dataframe
            The dataframe to evaluate the score and append the score column.
        """
        self.validate(check_model=True)
        channel = self.active_channel
        model = self.models[channel]
        train_variables = self.channel_config["train_variables"]
        train_data = df[train_variables]
        if self.data_preprocessor is not None:
            train_data = self.data_preprocessor(train_data)
        # use custom data format for an xgboost model
        if model.__class__.__module__.startswith("xgboost"):
            import xgboost as xgb
            train_data = xgb.DMatrix(train_data)
        score = model.predict(train_data)
        # check score quality
        score_min, score_max = np.min(score), np.max(score)
        if ((score_min < 0.) or (score_max > 1.)):
            raise RuntimeError("detected score value outside the range [0, 1]")
        
        # add score variable to dataframe
        ndim = np.ndim(score)
        num_class = self.channel_config["num_class"]

        channel_metadata = self.channel_metadata[channel]
        if num_class == 2:
            # require only output score to have dimension 1 in case of binary classification
            if (ndim != 1):
                raise RuntimeError(f"expect single output from binary classification but {ndim} received")
            score_column_name = self.get_score_column_name()
            df[score_column_name] = score
        else:
            for class_label, metadata in channel_metadata["classes"].items():
                score_column_name = self.get_score_column_name(class_label)
                class_index = metadata["encoded_index"]
                df[score_column_name] = score[:, class_index]
    
    def get_class_score_hist(self, class_label:Optional[str]=None,
                             cuts:Optional[str]=None) -> Dict[str, np.ndarray]:
        """
        Extract score distributions for the given MVA output class in the current channel
        
        Parameters
        ----------
        class_label: str
            name of the MVA output class from which the score distributions are extracted
        cuts: (optional) str
            additional cuts applied to each sample dataframes
        """
        self.validate(check_df=True)
        channel = self.active_channel
        num_class = self.channel_config["num_class"]
        if num_class == 2:
            if class_label:
                raise RuntimeError("class label is not needed for binary classification which gives only "
                                   "one score distribution")
        else:
            if not class_label:
                raise RuntimeError("class label must be given for multi-class MVA outputs to extract the "
                                   "corresponding score distributions")
        score_column_name = self.get_score_column_name(class_label)
        channel_metadata = self.channel_metadata[channel]
        scan_config = channel_metadata["classes"][class_label]["counting_significance"]
        n_bins = scan_config["n_bins"]
        bins = np.arange(0, 1, 1 / n_bins)
        signal_samples     = self.resolve_samples(scan_config["signal"])
        background_samples = self.resolve_samples(scan_config["background"])
        all_samples = self.resolve_samples(self.channel_config["test_samples"])
        all_hists   = {}
        for sample in all_samples:
            sample_df = self.channel_df[sample]
            if cuts is not None:
                sample_df =  sample_df.query(cuts,  engine="python")
            digits = np.digitize(sample_df[score_column_name], bins)
            group  = sample_df.groupby(digits)
            cat_weight_name = self.names["cat_weight"]
            bin_values  = group[cat_weight_name].sum().values
            bin_indices = group[cat_weight_name].sum().index.values - 1
            all_hists[sample] = np.zeros(n_bins)
            all_hists[sample][bin_indices] = bin_values
        signal_hists = np.zeros(n_bins)
        background_hists = np.zeros(n_bins)
        for sample in all_hists:
            if sample in signal_samples:
                signal_hists += all_hists[sample]
            elif sample in background_samples:
                background_hists += all_hists[sample]
        hists = {
            'all': all_hists,
            'signal': signal_hists,
            'background': background_hists
        }
        return hists
    
    def scan_class_bounds(self, class_score_hist, n_bins:int, n_boundaries:int,
                          minimum_yields:Optional[Dict]=None,
                          filter_regions:Optional[List[int]]=None) -> Dict:
        """
        Perform a boundary scan for the given MVA output class in the current channel
        
        Parameters
        ----------
        class_score_hist: dictionary
            A dictionary containing the histogram values for various type of samples.
        n_bins:
            Number of bins used for constructing the score histogram.
        n_boundaries:
            Number of score boundaries to use.
        minimum_yields: (optional) dictionary
            A dictionary with key as the sample that has a minimum yield requirement in
            all score regions formed from the score boundaries, and value as the value
            of the minimum yield.
        """
        all_hists = class_score_hist['all']
        signal_hists = class_score_hist['signal']
        background_hists = class_score_hist['background']
        if (n_bins != len(signal_hists)) or (n_bins != len(background_hists)):
            raise RuntimeError("number of bins requested must match the number of bins in the class score "
                               "histograms")
        with Timer() as t:
            boundary_indices  = self.get_boundary_indices(n_bins, n_boundaries)
            n_combinations = len(boundary_indices)
            self.stdout.info(f"INFO: Total number of boundary combinations: {n_combinations}")
            signal_yields     = self.get_region_yields(signal_hists, boundary_indices, filter_regions=filter_regions)
            background_yields = self.get_region_yields(background_hists, boundary_indices, filter_regions=filter_regions)
            # filter boundaries which the regions contain no signal or background yields
            valid_boundaries = np.all(background_yields != 0., axis=1) & np.all(signal_yields != 0., axis=1)
            # filter boundaries which the regions contain less than the minimum required yields for specific
            # samples such as the yy background
            if minimum_yields is not None:
                for sample in minimum_yields:
                    if sample not in all_hists:
                        raise RuntimeError(f"the sample {sample} is not used in testing")
                    min_yield = minimum_yields[sample]
                    sample_yields = self.get_region_yields(all_hists[sample], boundary_indices, filter_regions=filter_regions)
                    valid_boundaries &= np.all(sample_yields > min_yield, axis=1)
            self.stdout.info(f"INFO: Number of valid boundary combinations: {valid_boundaries.sum()}")
            if valid_boundaries.sum() == 0:
                self.stdout.info("INFO: Not enough combinations to scan due to minimum yield requirements. "
                                 "No boundaries are set.")
                return None
            combined_significances = get_combined_counting_significance(signal_yields[valid_boundaries],
                                                                        background_yields[valid_boundaries])
            bins = np.arange(0, 1, 1 / n_bins)
            max_boundary_idx = np.argmax(combined_significances)
            max_significance = combined_significances[max_boundary_idx]
            min_boundary_idx = np.argmin(combined_significances)
            min_significance = combined_significances[min_boundary_idx]
            if ((max_significance - min_significance) / min_significance) < 0.1:
                self.stdout.info("INFO: Improvement in Z is less than 10%. "
                                 "No boundaries are set.")
                return None
            max_bins         = boundary_indices[valid_boundaries][max_boundary_idx]
            max_boundary     = [bins[i] for i in max_bins]
        self.stdout.info(f"INFO: Boundary scan finished. Total time taken: {t.interval:.3f}s")
        result = {
            'boundaries'  : [round(v, 8) for v in max_boundary],
            'significance': max_significance 
        }
        return result
    
    def load_boundaries(self, filename:Optional[str]=None,
                        channel:Optional[str]=None) -> MultiClassBoundaryTree:
        """
        Load boundary information for a given channel.
        
        Parameters
        ----------
        filename: str
            Path to the json file containing the boundary data.
        channel: str
            The channel to which the boundary information is loaded.
            If None, the current channel is used.
        """
        channel = self._parse_channel(channel)
        if filename is None:
            filename = self.get_file("boundary_data", channel=channel)
        if not os.path.exists(filename):
            raise FileNotFoundError(f'boundary output "{filename}" does not exist')
        self.stdout.info(f'INFO: Cached boundary output for the channel {channel} from "{filename}"')
        boundary_tree = MultiClassBoundaryTree(score_base_name=self.names["score"])
        boundary_tree.load_tree(filename)
        self.boundary_trees[channel] = boundary_tree
        self.update_valid_categories(channel)
        self.update_summary(channel)
        return boundary_tree
                
    def scan_bounds(self, cache:bool=True) -> None:
        """
        Scan score boundaries for the current channel
        
        Parameters
        ----------
        cache: bool
            Whether to cache boundary data.
        """
        self.validate(check_df=True)
        channel = self.active_channel
        channel_config  = self.config['channels'][channel]
        boundary_tree = MultiClassBoundaryTree(score_base_name=self.names["score"])
        
        # caching
        if cache:
            # cache from existing data
            if (channel in self.boundary_trees) and (self.boundary_trees[channel]):
                return None
            # cache from boundary file
            elif self._has_file("boundary_data", channel=channel):
                filename = self.get_file("boundary_data", channel=channel)
                self.load_boundaries(filename)
                return None
            
        exclude_categories = channel_config["exclude_categories"]
        
        self.stdout.info("INFO: Scanning score boundaries...")
        # run over class
        class_labels = self.get_class_labels()
        class_metadata = self.channel_metadata[channel]["classes"]
        for depth, class_label in enumerate(class_labels):
            class_config = class_metadata[class_label]['counting_significance']
            cut_maps = boundary_tree.get_cut_maps()
            # run over preceding class regions
            for branch_index, cut_expr in cut_maps.items():
                if exclude_categories:
                    if depth == 0:
                        filter_regions = [category[depth] for category in exclude_categories]
                    else:
                        filter_regions = [category[depth] for category in exclude_categories \
                                          if category[:depth] == list(branch_index)]
                else:
                    filter_regions = None
                # multi-class case
                if class_label:
                    if branch_index is not None:
                        preceding_category_label = self.get_category_label(branch_index)
                        self.stdout.info(f"Class Name: {class_label} (region = {preceding_category_label})")
                    else:
                        self.stdout.info(f"Class Name: {class_label}")
                class_score_hist = self.get_class_score_hist(class_label, cut_expr)
                result = self.scan_class_bounds(class_score_hist,
                                                class_config['n_bins'],
                                                class_config['n_boundaries'],
                                                class_config['min_yield'],
                                                filter_regions=filter_regions)
                if result is not None:
                    # append the new boundaries to the preceding branch
                    boundary_tree.set_boundaries(class_label, **result, branch_index=branch_index)
                    self.stdout.info("\t boundaries: {}".format(result['boundaries']))
                    self.stdout.info("\t counting significance: {}".format(result['significance']))
        self.boundary_trees[channel] = boundary_tree
        self.update_valid_categories()
        self.update_summary()
    
    def update_valid_categories(self, channel:Optional[str]=None,
                                ignore_boundary_result:bool=False):
        """
        Update valid categories based on the boundary information from the current channel.
        
        Parameters
        ----------
        channel: (optional) str
            The channel from which the categories are retrieved.
            By default, the current channel is used.
        ignore_boundary_result: bool, default = False
            Ignore invalid boundaries from the boundary result. This will force all
            categories to be valid except those found in the exclude categories list. 
        """
        channel = self._parse_channel(channel)
        channel_config = self.channel_configs[channel]
        num_class = channel_config["num_class"]
        exclude_categories = channel_config["exclude_categories"]
        category_metadata = self.channel_metadata[channel]["categories"]
        boundary_tree = self.get_boundary_tree(channel)
        valid_category_indices = boundary_tree.get_branch_indices()
        if num_class == 2:
            for idx in valid_category_indices:
                if len(idx) != 1:
                    raise RuntimError(f'invalid category index {idx} from boundary tree: '
                                      f'expect index of size 1 for binary classification')
            for idx in exclude_categories:
                if not isinstance(idx, int) and (len(idx) != 1):
                    raise RuntimError(f'invalid category index {idx} from exclude categories: '
                                      f'expect index of size 1 for binary classification')
            valid_category_indices = [idx[0] for idx in valid_category_indices]
            exclude_categories = [idx[0] if not isinstance(idx, int) else idx for idx in exclude_categories]
        for category_index, category_data in category_metadata.items():
            if category_index in exclude_categories:
                category_data["valid"] = False
                continue
            if ignore_boundary_result or (category_index in valid_category_indices):
                category_data["valid"] = True
                
    def update_summary(self, channel:Optional[str]=None):
        """
        Update boundary and significance information to the category summary after boundaries are set.
        
        Parameters
        ----------
        channel: (optional) str
            The channel from which the summary information is updated.
            By default, the current channel is used.  
        """
        channel = self._parse_channel(channel)
        boundary_tree = self.get_boundary_tree(channel)
        
        boundary_summary = {}
        category_indices = self.get_category_indices(channel=channel)
        for category_index in category_indices:
            category_label = self.get_category_label(category_index, channel=channel)
            boundary_summary[category_label] = boundary_tree.get_combined_boundaries(category_index)
            
        significance_summary = boundary_tree.get_significance_summary()
        
        summary = {
            "boundary"    : boundary_summary,
            "significance": significance_summary
        }
        
        self.summary[channel] = summary
        
    def get_category_yields(self, resampling:bool=False,
                            resampling_random_state:Optional[int]=None):
        """
        Return category yields for the test samples in the current channel after category labels are applied.
        
        Parameters
        ----------
        resampling: bool, default = False
            Whether to resample the events (for bootstrapping).
        resampling_random_state: (optional) int
            The random state used to resample the events.        
        """
        self.validate(check_df=True, check_categories=True)
        channel = self.active_channel
        category_indices = self.get_category_indices()
        samples = self.channel_config['test_samples']
        weight_name = self.names["total_weight"]
        yields     = {}
        yields_err = {}
        for category_index in category_indices:
            category_name = self.get_category_name(category_index)
            yields[category_name]     = {}
            yields_err[category_name] = {}
            for sample in samples:
                is_data = self.is_data_sample(sample)
                do_blind = is_data and self.do_blind                
                category_df = self.get_category_df(sample, category_index,
                                                   do_blind=do_blind)
                key = sample
                if do_blind:
                    key = self.get_blind_sample_name(sample)
                if resampling:
                    category_df = self.get_resampled_df(category_df, random_state=resampling_random_state)
                yields[category_name][key] = float(category_df[weight_name].sum())
                if is_data:
                    #calculate poisson interval
                    poisson_interval = get_poisson_interval([yields[category_name][key]])
                    yields_err[category_name][key] = {"errlo": poisson_interval["lo"][0],
                                                      "errhi": poisson_interval["hi"][0]}
                else:
                    err_val = np.sqrt((category_df[weight_name] ** 2).sum())
                    yields_err[category_name][key] = {"errlo": err_val, "errhi": err_val}
        return yields, yields_err
    
    def update_category_yields(self):
        """
        Update category yields for the test samples in the current channel after category labels are applied.
        """        
        self.stdout.info(f"INFO: Evaluating sample yields")
        yields, yields_err = self.get_category_yields()
        channel = self.active_channel
        self.category_yields[channel]     = yields
        self.category_yields_err[channel] = yields_err
        
    def apply_category(self, df:pd.DataFrame):
        """
        Apply category label to a given sample dataframe according to the score boundaries.
        
        Parameters
        ----------
        df: pandas.DataFrame
            Sample dataframe to apply to category label.
        """
        self.validate(check_boundaries=True)
        channel       = self.active_channel
        channel_index = self.target_channels.index(channel)
        num_class     = self.channel_config['num_class']
        boundary_tree = self.get_boundary_tree()
        if num_class == 2:
            boundaries = boundary_tree.get_boundaries()
            score_column_name = self.get_score_column_name()
            cattegory_column_name = self.get_category_column_name()
            categories = 1000 * channel_index + np.digitize(df[score_column_name], boundaries)
            df[cattegory_column_name] = categories
        else:
            cut_maps = boundary_tree.get_cut_maps()
            class_labels = self.get_class_labels()
            category_column_names = [self.get_category_column_name(class_label) for class_label in class_labels]
            for category_index, cut_expr in cut_maps.items():
                masks = df.query(cut_expr).index
                columns, indices = self.pair_category_columns(category_index)
                df.loc[masks, columns]  = indices
                df.loc[masks, columns] += 1000 * channel_index
            df.fillna(-1, inplace=True)

    def apply_categories(self):
        """
        Apply category label to all sample dataframes for the current channel.
        """
        self.validate(check_df=True, check_boundaries=True)
        for sample, sample_df in self.channel_df.items():
            self.apply_category(sample_df)
    
    def pair_category_columns(self, category_index:Union[int, List[int], Tuple[int]]) -> Tuple:
        # index for binary classification
        if isinstance(category_index, int):
            return ([self.get_category_column_name()], [category_index])
        else:
            class_labels = self.get_class_labels()
            category_column_names = [self.get_category_column_name(class_label) for class_label in class_labels]
            return category_column_names[:len(category_index)], list(category_index)
    
    def get_category_query_str(self, category_or_index:Union[str, int, List[int], Tuple[int]]) -> str:
        """
        Parameters
        ----------
        category_or_index: str / int / list of int / tuple of int
            The category name or category index to perform the query.
            If string, it is interpretated as a category name (without the channel prefix).
            If any of (int, list of int, tuple of int), it is interpreted as the category
            index.
        """
        if isinstance(category_or_index, str):
            category_index = self.get_category_index(category_or_index)
        else:
            category_index = category_or_index
        
        components = []
        columns, indices = self.pair_category_columns(category_index)
        for column, index in zip(columns, indices):
            components.append(f"({column} % 1000 == {index})")
        query_str = " & ".join(components)
        return query_str
    
    def get_category_df(self, sample_or_df:Union[str, pd.DataFrame],
                        category_or_index:Optional[Union[str, int, List[int], Tuple[int]]]=None,
                        resampling:bool=False,
                        resampling_random_state:Optional[int]=None,
                        columns:Optional[List[str]]=None,
                        selection:Optional[str]=None,
                        do_blind:bool=False):
        """
        Return dataframe from a given category.
            
        Parameters
        ----------
        sample_or_df: str / pandas.DataFrame
            If str, it is the name of the sample loaded in the current channel.
            If pandas.DataFrame, it is a custom dataframe with columns containing
            the proper category indices.
        category_or_index: str / int / list of int / tuple of int
            The category name or category index to perform the query.
            If string, it is interpretated as a category name (without the channel prefix).
            If any of (int, list of int, tuple of int), it is interpreted as the category
            index.
        resampling: bool, default = False
            Whether to resample the events (for bootstrapping).
        resampling_random_state: (optional) int
            The random state used to resample the events.
        columns: (optional) list of str
            Columns to keep in the output dataframe.
        selection: (optional) str
            Additional selection applied to the dataframe.
        do_blind: bool, default = False
            Whether to apply a blind selection.
        """
        if isinstance(sample_or_df, str):
            if category_or_index in self.target_channels:
                self.validate(check_df=True)
            else:
                self.validate(check_df=True, check_boundaries=True, check_categories=True)
            df = self.channel_df[sample_or_df]
        else:
            df = sample_or_df
        if category_or_index is not None:
            if category_or_index not in self.target_channels:
                query_str = self.get_category_query_str(category_or_index)
                df = df.query(query_str)
        if selection is not None:
            df = df.query(selection)
        if do_blind:
            df = self.get_blind_df(df)
        if resampling:
            df = self.get_resampled_df(df, random_state=resampling_random_state)
        if columns is not None:
            available_columns = list(df.columns)
            valid_columns = [c for c in columns if c in available_columns]
            missing_columns = [c for c in columns if c not in available_columns]
            if len(missing_columns) > 0:
                self.stdout.debug(f'DEBUG: Found missing columns in the datafarme: {", ".join(missing_columns)}')
            df = df[valid_columns]
        return df
    
    def get_resampled_df(self, df:pd.DataFrame,
                         random_state:Optional[int]=None):
        """
        Return dataframe with resampled events (for bootstrapping).
            
        Parameters
        ----------
        df: pandas.DataFrame
            Target dataframe.
        random_state: (optional) int
            The random state used to resample the events.
        """        
        df = df[df[self.names["total_weight"]] != 0]
        df = df.sample(n=df.shape[0], replace=True, random_state=random_state)
        return df
    
    def get_blind_df(self, df:pd.DataFrame) -> pd.DataFrame:
        """
        Return dataframe with events in the signal region removed.
        
        Parameters
        ----------
        df: pandas.DataFrame
            Target dataframe.
        """
        blind_selection = self.get_blind_selection()
        if blind_selection not in ["1", None]:
            df = df.query(blind_selection)
        return df
    
    def _get_sumed_sample_yields(self, samples:List[str],
                                 category_or_index:Union[str, int, List[int], Tuple[int]],
                                 weight_name:str):
        yields = 0.
        for sample in samples:
            category_df = self.get_category_df(sample, category_or_index)
            yields += category_df[weight_name].sum()
        return yields
    
    def evaluate_benchmark_significance(self):
        self.validate(check_df=True, check_boundaries=True, check_categories=True)
        if ('benchmark' not in self.config) or ('counting_significance' not in self.config['benchmark']):
            self.stdout.info("INFO: No significance benchmark defined. Skipped evaluation.")
        benchmark_config = self.config['benchmark']['counting_significance']
        benchmarks = list(benchmark_config)
        benchmark_significance = {}
        channel = self.active_channel
        category_labels = self.get_category_labels(valid_only=True)
        for benchmark in benchmarks:
            benchmark_significance[benchmark] = {}
            signal_samples     = self.resolve_samples(benchmark_config[benchmark]['signal'])
            background_samples = self.resolve_samples(benchmark_config[benchmark]['background'])
            Z = []
            for category_label in category_labels:
                yield_s = self._get_sumed_sample_yields(signal_samples, category_label, self.names["cat_weight"])
                yield_b = self._get_sumed_sample_yields(background_samples, category_label, self.names["cat_weight"])
                Z_cat   = get_counting_significance(yield_s, yield_b)
                category_name = f"{channel}_{category_label}"
                benchmark_significance[benchmark][category_name] = Z_cat
                Z.append(Z_cat)
            combined_Z = float(np.sqrt(np.sum(np.array(Z) ** 2)))
            benchmark_significance[benchmark][f"{channel}_combined"] = combined_Z
        self.benchmark_significance = combine_dict(self.benchmark_significance, benchmark_significance)
        
    def update_combined_benchmark_significance(self):
        self.validate(check_channels=True)
        category_names = self.get_all_category_names()
        benchmarks = list(self.config['benchmark']['counting_significance'])
        for benchmark in benchmarks:
            if benchmark not in self.benchmark_significance:
                raise RuntimeError(f'missing benchmark information for the benchmark "{benchmark}"')
            benchmark_data = self.benchmark_significance[benchmark]
            Z = []
            for category_name in category_names:
                if category_name not in benchmark_data:
                    raise RuntimeError(f'missing benchmark information for the category "{category_name}"')
                Z.append(benchmark_data[category_name])
            combined_Z = float(np.sqrt(np.sum(np.array(Z) ** 2)))
            benchmark_data['combined'] = combined_Z
    
    def run_channel_categorization(self, channel:str, cache_boundaries:bool=True,
                                   event_indices:Optional[Dict[str, np.ndarray]]=None):
        """
        Perform event categorization for a given channel.
            
        Parameters
        ----------
        channel: str
           Name of analysis channel.
        cache: bool, default = True
           Cache categorization result.
        cat_event_number: (optional) dict
           If specified, event categorization will be performed on a specific set of events for each sample
           given in the dictionary. The format is { sample: array of event numbers }.
        """
        if event_indices is not None:
            self.stdout.info("INFO: Categorization will be done on a given set of event numbers for each sample")
        self.set_active_channel(channel)
        self.load_channel_df(event_indices=event_indices)
        self.scan_bounds(cache=cache_boundaries)
        self.apply_categories()
        self.evaluate_benchmark_significance()
    
    def save_dataframe_as_array(self, df:pd.DataFrame, savepath:str, fmt:Optional[str]=None):
        """
        Save dataframe in array (csv/ht) format.

        Parameters
        ----------
        df: pandas.DataFrame
            Dataframe to save.
        savepath: string
            Path to save the output.       
        """
        if fmt is None:
            fmt = self.get_analysis_data_format()
        if fmt == "csv":
            df.to_csv(savepath, mode="w", index=False)
        elif fmt == "h5":
            df.to_hdf(outpath, key="categorized_data", mode="w",
                      index=False, complevel=None)
        
    def save_dataframe_as_histogram(self, df:pd.DataFrame, savepath:str):
        """
        Save dataframe in root histogram format.

        Parameters
        ----------
        df: pandas.DataFrame
            Dataframe to save.
        savepath: string
            Path to save the output.       
        """
        from quickstats.components import ExtendedRFile
        kwargs = {
            "name"   : self.config["observable"]["name"],
            "n_bins" : self.config["observable"]["n_bins"],
            "xmin"   : self.config["observable"]["bin_range"][0],
            "xmax"   : self.config["observable"]["bin_range"][1],
            "column" : self.config["observable"]["name"],
            "weight" : self.names["total_weight"]
        }
        rf = ExtendedRFile(df[[kwargs["column"], kwargs["weight"]]])
        hist = rf.get_Histo1D(**kwargs)
        # remove bins with negative weights
        from quickstats.interface.root import TH1
        neg_bin_indices, neg_bin_values = TH1.remove_negative_bins(hist)
        if len(neg_bin_indices) > 0:
            index_str = [f"{i}" for i in (neg_bin_indices + 1)]
            value_str = [f"{v}" for v in neg_bin_values]
            self.stdout.warning(f'WARNING: Detected bins with negative weights '
                                f'(bins = {", ".join(index_str)}, '
                                f'weights = {", ".join(value_str)}). '
                                f'Bin weight will be set to 0.')
        rf.save_components(savepath, components=[hist])
        
    def save_dataframe_as_minitree(self, df:pd.DataFrame, savepath:str):
        """
        Save dataframe in root minitree format.

        Parameters
        ----------
        df: pandas.DataFrame
            Dataframe to save.
        savepath: string
            Path to save the output.       
        """
        from quickstats.utils.data_conversion import dataframe2root
        dataframe2root(df, savepath, self.treename)
            
    def save_channel_summary(self, channel:Optional[str]=None) -> None:
        """
        Save category summary and boundary data for a given channel.
        
        Parameters
        ----------
        channel: (optional) str
            The channel from which the category summary and boundary data are retrieved.
            By default, the current channel is used.
        """
        channel = self._parse_channel(channel)
        # save category summary
        savepath = self.get_file("category_summary", channel=channel)
        if channel not in self.summary:
            raise RuntimeError(f'category summary for the chanenl "{channel}" not initialized')
        with open(savepath, "w") as outfile:
            json.dump(self.summary[channel], outfile, indent=2)
        self.stdout.info(f'INFO: Saved category summary to {savepath}')
        # save boundary data
        savepath = self.get_file("boundary_data", channel=channel)
        boundary_tree = self.get_boundary_tree(channel=channel)
        with open(savepath, "w") as outfile:
            json.dump(boundary_tree.tree, outfile, indent=2)
        self.stdout.info(f'INFO: Saved boundary data to {savepath}')
        
    def save_channel_yields(self, channel:Optional[str]=None,
                            resampling:bool=False,
                            resampling_random_state:Optional[int]=None) -> None:
        """
        Save yield information for a given channel.
        
        Parameters
        ----------
        channel: (optional) str
            The channel from which the yield information is retrieved.
            By default, the current channel is used.
        resampling: bool, default = False
            Whether to resample the events (for bootstrapping).
        resampling_random_state: (optional) int
            The random state used to resample the events.                
        """
        channel = self._parse_channel(channel)
        if (channel not in self.category_yields) or (channel not in self.category_yields_err):
            raise RuntimeError(f'yield information for the channel "{channel}" not initialized')
        if resampling and (channel != self.active_channel):
            raise RuntimeError('resampling can only be performed on the active channel')
        if resampling:
            yields, yields_err = self.get_category_yields(resampling=resampling,
                                                          resampling_random_state=resampling_random_state)
        else:
            yields     = self.category_yields[channel]
            yields_err = self.category_yields_err[channel]
        for category, category_yields in yields.items():
            savepath = self.get_file("yield_data", category=category)
            with open(savepath, "w") as outfile:
                json.dump(category_yields, outfile, indent=2)
            self.stdout.info(f'INFO: Saved yield information to {savepath}')
        for category, category_yields_err in yields_err.items():
            savepath = self.get_file("yield_err_data", category=category)
            with open(savepath, "w") as outfile:
                json.dump(category_yields_err, outfile, indent=2)
            self.stdout.info(f'INFO: Saved yield uncertainty information to {savepath}')
            
    def _get_minimal_sample_outputs(self):
        """Get minimal set of output samples to be saved; analysis specific
        """
        ARRAY     = CategorizationOutputFormat.ARRAY
        MINITREE  = CategorizationOutputFormat.MINITREE
        HISTOGRAM = CategorizationOutputFormat.HISTOGRAM        
        minimal_samples = {
            ARRAY     : [],
            MINITREE  : [],
            HISTOGRAM : []
        }
        return minimal_samples
    
    def _get_standard_sample_outputs(self):
        """Get standard set of output samples to be saved; analysis specific
        """
        ARRAY     = CategorizationOutputFormat.ARRAY
        MINITREE  = CategorizationOutputFormat.MINITREE
        HISTOGRAM = CategorizationOutputFormat.HISTOGRAM        
        minimal_samples = {
            ARRAY     : [],
            MINITREE  : [],
            HISTOGRAM : []
        }
        return minimal_samples
    
    def _get_required_sample_outputs(self, mode:Union[str, CategorizationOutputMode]="minimal"):
        """
        Return the list of samples to be saved in array (csv/h5), histogram and minitree formats 
        in the current channel.

        The full sample list is taken from "test_samples" of a channel-level configuration.

        Parameters
        ----------
        mode: str or CategorizationOutputMode
            Output mode. Available choices are
                "minimal"  : custom minimal output mode defined by analysis
                "minitree" : save minitrees for all samples
                "array"    : save array (csv/h5) for all samples
                "standard" : custom standard output mode defined by analysis
                "full"     : save array (csv/h5), histograms and minitrees for all samples
        """
        self.validate()
        all_samples = list(self.channel_config['test_samples'])
        
        resolved_mode = CategorizationOutputMode.parse(mode)
        ARRAY     = CategorizationOutputFormat.ARRAY
        MINITREE  = CategorizationOutputFormat.MINITREE
        HISTOGRAM = CategorizationOutputFormat.HISTOGRAM
        sample_outputs = {
            ARRAY     : [],
            MINITREE  : [],
            HISTOGRAM : []
        }
        if resolved_mode == CategorizationOutputMode.MINIMAL:
            sample_outputs = self._get_minimal_sample_outputs()
        elif resolved_mode == CategorizationOutputMode.MINITREE:
            sample_outputs[MINITREE] = all_samples
        elif resolved_mode == CategorizationOutputMode.ARRAY:
            sample_outputs[ARRAY] = all_samples
        elif resolved_mode == CategorizationOutputMode.STANDARD:
            sample_outputs[ARRAY]     = all_samples
            sample_outputs[MINITREE]  = all_samples            
        elif resolved_mode == CategorizationOutputMode.FULL:
            sample_outputs[ARRAY]     = all_samples
            sample_outputs[HISTOGRAM] = all_samples
            sample_outputs[MINITREE]  = all_samples
        else:
            raise ValueError(f'unknown output mode "{mode}"')
        return sample_outputs
                
    def save_channel_outputs(self, cache:bool=True, mode:str="minimal",
                             save_yield:bool=True,
                             save_summary:bool=True,
                             resampling:bool=False,
                             resampling_random_state:Optional[int]=None,
                             **kwargs):
        """
        Save collection of category outputs for the current channel.
        
        Parameters
        ----------
        resampling: bool, default = False
            Whether to resample the events (for bootstrapping).
        resampling_random_state: (optional) int
            The random state used to resample the events.                
        """        
        
        self.validate(check_df=True, check_boundaries=True, check_categories=True)
        
        channel = self.active_channel
        columns = self.get_columns_to_save()
        
        category_indices = self.get_category_indices()
        samples = self.channel_config['test_samples']
        sample_outputs = self._get_required_sample_outputs(mode=mode)
        self.check_missing_columns(columns)
        
        if resampling:
            self.stdout.info("NOTE: Resampling mode is enabled. The following outputs will "
                             "be saved with resampled events: csv, minitree, histogram, "
                             "yield, data point.")
        save_methods = {
            CategorizationOutputFormat.ARRAY     : self.save_dataframe_as_array,
            CategorizationOutputFormat.MINITREE  : self.save_dataframe_as_minitree,
            CategorizationOutputFormat.HISTOGRAM : self.save_dataframe_as_histogram,
        }

        for output_format in sample_outputs:
            self.path_manager.makedirs([output_format.dirname])
            
        fmt = self.get_analysis_data_format() 
        # also save the channel-level df
        category_indices.append(None)
        
        for sample in samples:
            for category_index in category_indices:
                if category_index is None:
                    category_name = channel
                else:
                    category_name = self.get_category_name(category_index)
                do_blind = self.is_data_sample(sample) and self.do_blind
                df = self.get_category_df(sample, category_index,
                                          columns=columns,
                                          do_blind=do_blind,
                                          resampling=resampling,
                                          resampling_random_state=resampling_random_state)
                for output_format in CategorizationOutputFormat:
                    if sample not in sample_outputs[output_format]:
                        continue
                    savepath = self.get_file(output_format.filename, sample=sample,
                                             category=category_name, fmt=fmt)
                    if cache and os.path.exists(savepath):
                        self.stdout.info(f'INFO: Cached {output_format.short_name} output from {savepath}')
                        continue
                    save_methods[output_format](df, savepath)
                    self.stdout.info(f'INFO: Saved {output_format.short_name} output to {savepath}')
                    
        if save_yield:
            self.update_category_yields()
            self.path_manager.makedirs(["yield"])
            self.save_channel_yields(channel,
                                     resampling=resampling,
                                     resampling_random_state=resampling_random_state)
            
        if save_summary:
            self.path_manager.makedirs(["summary"])
            self.save_channel_summary(channel)
    
    def check_missing_columns(self, columns:List[str]):
        self.validate(check_df=True)
        
        missing_column_by_sample = {}
        samples = self.channel_config['test_samples']
        for sample in samples:
            missing_columns = list(set(columns) - set(self.channel_df[sample].columns))
            if len(missing_columns) > 0:
                missing_column_by_sample[sample] = missing_columns
        if len(missing_column_by_sample) > 0:
            missing_same_columns = len(set(tuple(cols) for cols in missing_column_by_sample.values())) == 1
            all_samples_missing = set(missing_column_by_sample.keys()) == set(self.channel_df.keys())
            if missing_same_columns:
                columns = list(missing_column_by_sample.values())[0]
                if all_samples_missing:
                    self.stdout.warning(f"WARNING: The following save columns are missing for all samples: "
                                        f"{', '.join(columns)}")
                else:
                    samples = list(missing_column_by_sample.keys())
                    self.stdout.warning(f"WARNING: The following save columns are missing for the samples "
                                        f"{', '.join(samples)}: {', '.join(columns)}")
            else:
                for sample, columns in missing_column_by_sample.items():
                    self.stdout.warning(f"WARNING: The following save columns are missing for the sample {sample}: "
                                        f"{', '.join(columns)}")
    
    def get_yield_files_and_description(self):
        result = {
            "yield_data"     : "yield information",
            "yield_err_data" : "yield uncertainty information"
        }
        return result
    
    def save_combined_yields(self, cache:bool=True):
        self.validate(check_channels=True)
        category_names = self.get_all_category_names()
        yield_info = self.get_yield_files_and_description()
        for yield_file, description in yield_info.items():
            merged_yield_file = f"merged_{yield_file}"
            savepath = self.get_file(merged_yield_file)
            if cache and os.path.exists(savepath):
                self.stdout.info(f"INFO: Cached combined {description} from {savepath}")
            yield_data = {}
            for category_name in category_names:
                filepath = self.get_file(yield_file, validate=True, category=category_name)
                with open(filepath, "r") as file:
                    yield_data[category_name] = json.load(file)
            with open(savepath, "w") as outfile:
                json.dump(yield_data, outfile, indent=2)
                self.stdout.info(f"INFO: Saved combined {description} to {savepath}")
                
    def save_combined_benchmark_significance(self, cache:bool=True):
        self.validate(check_channels=True)
        benchmark_savepath = self.get_file("benchmark_significance")
        if (is_valid_file(benchmark_savepath)) and cache:
            self.stdout.info(f"INFO: Cached benchmark significance data from {benchmark_savepath}")
            if not self.benchmark_significance:
                with open(benchmark_savepath, "r") as file:
                    self.benchmark_significance = json.load(file)
        elif self.benchmark_significance:
            with open(benchmark_savepath, "w") as file:
                json.dump(self.benchmark_significance, file)
                self.stdout.info(f"INFO: Saved benchmark significance data to {benchmark_savepath}")
                      
    def merge_category_outputs(self, save_yield:bool=True,
                               save_benchmark:bool=True,
                               cache:bool=True):
        for channel in self.target_channels:
            if channel not in self.boundary_trees:
                self.load_boundaries(channel=channel)
        
        self.stdout.info("INFO: Merging outputs from all channels...")
       
        if save_yield:
            self.save_combined_yields(cache=cache)

        # merge benchmark significance
        if self.benchmark_significance:
            self.update_combined_benchmark_significance()
            
        if save_benchmark:
            self.save_combined_benchmark_significance(cache=cache)
    
    def load_cached_category_df(self, category:str, samples:Optional[List[str]]=None):
        if category in self.target_channels:
            channel = category
        else:
            channel = self.get_category_channel(category)
        if samples is None:
            resolved_samples = self.channel_configs[channel]['test_samples']
        elif isinstance(samples, str):
            resolved_samples = [samples]
        else:
            resolved_samples = samples
        
        dfs = {}
        fmt = self.get_analysis_data_format()
        for sample in resolved_samples:
            sample_path = self.get_file("categorized_array_sample", sample=sample,
                                        category=category, fmt=fmt)
            if not os.path.exists(sample_path):
                raise FileNotFoundError(f'missing categorized array file for the sample "{sample}" '
                                        f'({sample_path})')
            df = DataLoader._read_data_from_file(sample_path)
            if self.is_data_sample(sample) and self.do_blind:
                df = self.get_blind_df(df)
            dfs[sample] = df
        if isinstance(samples, str):
            return dfs[samples]
        return dfs

    @staticmethod
    def get_boundary_indices(n_bins:int, n_cut:int=2):
        n = n_bins
        k = n_cut
        a = np.ones((k, n-k+1), dtype="int32")
        a[0] = np.arange(n-k+1)
        for j in range(1, k):
            reps = (n-k+j) - a[j-1]
            a = np.repeat(a, reps, axis=1)
            ind = np.add.accumulate(reps)
            a[j, ind[:-1]] = 1-reps[1:]
            a[j, 0] = j
            a[j] = np.add.accumulate(a[j])
        a = a.T
        mask = (a[:, 0] != 0)
        return a[mask]

    @staticmethod
    def get_region_yields(data:np.ndarray, boundary_indices:np.ndarray,
                          filter_regions:Optional[List[int]]=None):
        # much faster implementation
        n = data.shape[0]
        d1, d2 = boundary_indices.shape[0], boundary_indices.shape[1]
        slice_indices = np.insert(boundary_indices, [0, d2], [0, n], axis=1).flatten()
        new_data = np.insert(data, n, 0)
        result = np.add.reduceat(new_data, slice_indices).reshape(d1, d2 + 2)
        result = np.delete(result, -1, axis=1)
        # remove yields from disgarded region(s)
        if filter_regions is not None:
            result = np.delete(result, filter_regions, axis=1)
        return result